# Stubs for cachecontrol.controller (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

# from .cache import DictCache as DictCache
# from .serialize import Serializer as Serializer

logger = ...  # type: Any
URI = ...  # type: Any

def parse_uri(uri): ...

class CacheController:
    cache = ...  # type: Any
    cache_etags = ...  # type: Any
    serializer = ...  # type: Any
    def __init__(self, cache=None, cache_etags=True, serializer=None): ...
    @classmethod
    def cache_url(cls, uri): ...
    def parse_cache_control(self, headers): ...
    def cached_request(self, request): ...
    def conditional_headers(self, request): ...
    def cache_response(self, request, response, body=None): ...
    def update_cached_response(self, request, response): ...



from typing import Any

def list2set(seq): ...
def first(seq): ...
def uniq(sequence, strip: int = ...): ...
def more_than(sequence, number): ...
def to_term(s, default: Any | None = ...): ...
def from_n3(
    s: str, default: Any | None = ..., backend: Any | None = ..., nsm: Any | None = ...
): ...
def check_context(c) -> None: ...
def check_subject(s) -> None: ...
def check_predicate(p) -> None: ...
def check_object(o) -> None: ...
def check_statement(triple) -> None: ...
def check_pattern(triple) -> None: ...
def date_time(t: Any | None = ..., local_time_zone: bool = ...): ...
def parse_date_time(val): ...
def guess_format(fpath, fmap: Any | None = ...): ...
def find_roots(graph, prop, roots: Any | None = ...): ...
def get_tree(
    graph,
    root,
    prop,
    mapper=...,
    sortkey: Any | None = ...,
    done: Any | None = ...,
    dir: str = ...,
): ...



"""Python code generator for a given schema salad definition."""
import textwrap
from io import StringIO
from typing import (
    IO,
    Any,
    Dict,
    List,
    MutableMapping,
    MutableSequence,
    Optional,
    Set,
    Union,
)

try:
    import black
except ModuleNotFoundError:
    black = None  # type: ignore[assignment]

from pkg_resources import resource_stream

from . import schema
from .codegen_base import CodeGenBase, TypeDef
from .exceptions import SchemaException
from .schema import shortname

_string_type_def = TypeDef("strtype", "_PrimitiveLoader(str)")
_int_type_def = TypeDef("inttype", "_PrimitiveLoader(int)")
_float_type_def = TypeDef("floattype", "_PrimitiveLoader(float)")
_bool_type_def = TypeDef("booltype", "_PrimitiveLoader(bool)")
_null_type_def = TypeDef("None_type", "_PrimitiveLoader(type(None))")
_any_type_def = TypeDef("Any_type", "_AnyLoader()")

prims = {
    "http://www.w3.org/2001/XMLSchema#string": _string_type_def,
    "http://www.w3.org/2001/XMLSchema#int": _int_type_def,
    "http://www.w3.org/2001/XMLSchema#long": _int_type_def,
    "http://www.w3.org/2001/XMLSchema#float": _float_type_def,
    "http://www.w3.org/2001/XMLSchema#double": _float_type_def,
    "http://www.w3.org/2001/XMLSchema#boolean": _bool_type_def,
    "https://w3id.org/cwl/salad#null": _null_type_def,
    "https://w3id.org/cwl/salad#Any": _any_type_def,
    "string": _string_type_def,
    "int": _int_type_def,
    "long": _int_type_def,
    "float": _float_type_def,
    "double": _float_type_def,
    "boolean": _bool_type_def,
    "null": _null_type_def,
    "Any": _any_type_def,
}


def fmt(text: str, indent: int) -> str:
    """
    Use black to format this snippet.

    :param indent the indent level for the current context
    """
    if not black:
        raise Exception(
            "Must install 'black' to use the Python code generator. "
            "If installing schema-salad via pip, try `pip install schema-salad[pycodegen]`."
        )
    return textwrap.indent(
        black.format_str(
            text,
            mode=black.Mode(
                target_versions={black.TargetVersion.PY36}, line_length=88 - indent
            ),
        ),
        " " * indent,
    )


class PythonCodeGen(CodeGenBase):
    """Generation of Python code for a given Schema Salad definition."""

    def __init__(
        self,
        out: IO[str],
        copyright: Optional[str],
        parser_info: str,
    ) -> None:
        super().__init__()
        self.out = out
        self.current_class_is_abstract = False
        self.serializer = StringIO()
        self.idfield = ""
        self.copyright = copyright
        self.parser_info = parser_info

    @staticmethod
    def safe_name(name: str) -> str:
        """Generate a safe version of the given name."""
        avn = schema.avro_field_name(name)
        if avn.startswith("anon."):
            avn = avn[5:]
        if avn in ("class", "in"):
            # reserved words
            avn = avn + "_"
        return avn

    def prologue(self) -> None:
        """Trigger to generate the prolouge code."""
        self.out.write(
            """#
# This file was autogenerated using schema-salad-tool --codegen=python
# The code itself is released under the Apache 2.0 license and the help text is
# subject to the license of the original schema.
"""
        )
        if self.copyright:
            self.out.write(
                """
#
# The original schema is {copyright}.
""".format(
                    copyright=self.copyright
                )
            )

        stream = resource_stream(__name__, "python_codegen_support.py")
        python_codegen_support = stream.read().decode("UTF-8")
        self.out.write(python_codegen_support[python_codegen_support.find("\n") + 1 :])
        stream.close()
        self.out.write("\n\n")

        self.out.write(
            f"""def parser_info() -> str:
    return "{self.parser_info}"


"""
        )

        for primative in prims.values():
            self.declare_type(primative)

    def begin_class(
        self,  # pylint: disable=too-many-arguments
        classname: str,
        extends: MutableSequence[str],
        doc: str,
        abstract: bool,
        field_names: MutableSequence[str],
        idfield: str,
        optional_fields: Set[str],
    ) -> None:
        classname = self.safe_name(classname)

        if extends:
            ext = ", ".join(self.safe_name(e) for e in extends)
        else:
            ext = "Savable"

        self.out.write(fmt(f"class {classname}({ext}):\n    pass", 0)[:-9])
        # make a valid class for Black, but then trim off the "pass"

        if doc:
            self.out.write(fmt(f'"""\n{doc}\n"""\n', 4) + "\n")

        self.serializer = StringIO()

        self.current_class_is_abstract = abstract
        if self.current_class_is_abstract:
            self.out.write("    pass\n\n\n")
            return

        required_field_names = [f for f in field_names if f not in optional_fields]
        optional_field_names = [f for f in field_names if f in optional_fields]

        safe_inits: List[str] = ["        self,"]
        safe_inits.extend(
            [
                f"        {self.safe_name(f)}: Any,"
                for f in required_field_names
                if f != "class"
            ]
        )
        safe_inits.extend(
            [
                f"        {self.safe_name(f)}: Optional[Any] = None,"
                for f in optional_field_names
                if f != "class"
            ]
        )
        self.out.write(
            "    def __init__(\n"
            + "\n".join(safe_inits)
            + "\n        extension_fields: Optional[Dict[str, Any]] = None,"
            + "\n        loadingOptions: Optional[LoadingOptions] = None,"
            + "\n    ) -> None:\n"
            + """
        if extension_fields:
            self.extension_fields = extension_fields
        else:
            self.extension_fields = CommentedMap()
        if loadingOptions:
            self.loadingOptions = loadingOptions
        else:
            self.loadingOptions = LoadingOptions()
"""
        )
        field_inits = ""
        for name in field_names:
            if name == "class":
                field_inits += """        self.class_ = "{}"
""".format(
                    classname
                )
            else:
                field_inits += """        self.{0} = {0}
""".format(
                    self.safe_name(name)
                )
        self.out.write(
            field_inits
            + f"""
    @classmethod
    def fromDoc(
        cls,
        doc: Any,
        baseuri: str,
        loadingOptions: LoadingOptions,
        docRoot: Optional[str] = None,
    ) -> "{classname}":
        _doc = copy.copy(doc)
        if hasattr(doc, "lc"):
            _doc.lc.data = doc.lc.data
            _doc.lc.filename = doc.lc.filename
        _errors__ = []
"""
        )

        self.idfield = idfield

        self.serializer.write(
            """
    def save(
        self, top: bool = False, base_url: str = "", relative_uris: bool = True
    ) -> Dict[str, Any]:
        r: Dict[str, Any] = {}
        for ef in self.extension_fields:
            r[prefix_url(ef, self.loadingOptions.vocab)] = self.extension_fields[ef]
"""
        )

        if "class" in field_names:
            self.out.write(
                """
        if _doc.get("class") != "{class_}":
            raise ValidationException("Not a {class_}")

""".format(
                    class_=classname
                )
            )

            self.serializer.write(
                """
        r["class"] = "{class_}"
""".format(
                    class_=classname
                )
            )

    def end_class(self, classname: str, field_names: List[str]) -> None:
        """Signal that we are done with this class."""
        if self.current_class_is_abstract:
            return

        self.out.write(
            fmt(
                """
extension_fields: Dict[str, Any] = {{}}
for k in _doc.keys():
    if k not in cls.attrs:
        if ":" in k:
            ex = expand_url(
                k, "", loadingOptions, scoped_id=False, vocab_term=False
            )
            extension_fields[ex] = _doc[k]
        else:
            _errors__.append(
                ValidationException(
                    "invalid field `{{}}`, expected one of: {attrstr}".format(
                        k
                    ),
                    SourceLine(_doc, k, str),
                )
            )
            break

if _errors__:
    raise ValidationException(\"Trying '{class_}'\", None, _errors__)
""".format(
                    attrstr=", ".join([f"`{f}`" for f in field_names]),
                    class_=self.safe_name(classname),
                ),
                8,
            )
        )

        self.serializer.write(
            """
        # top refers to the directory level
        if top:
            if self.loadingOptions.namespaces:
                r["$namespaces"] = self.loadingOptions.namespaces
            if self.loadingOptions.schemas:
                r["$schemas"] = self.loadingOptions.schemas
"""
        )

        self.serializer.write("        return r\n\n")

        self.serializer.write(
            fmt(f"""attrs = frozenset(["{'", "'.join(field_names)}"])\n""", 4)
        )

        safe_init_fields = [
            self.safe_name(f) for f in field_names if f != "class"
        ]  # type: List[str]

        safe_inits = [f + "=" + f for f in safe_init_fields]

        safe_inits.extend(
            ["extension_fields=extension_fields", "loadingOptions=loadingOptions"]
        )

        self.out.write(
            "        return cls(\n            "
            + ",\n            ".join(safe_inits)
            + ",\n        )\n"
        )

        self.out.write(str(self.serializer.getvalue()))

        self.out.write("\n\n")

    def type_loader(
        self, type_declaration: Union[List[Any], Dict[str, Any], str]
    ) -> TypeDef:
        """Parse the given type declaration and declare its components."""
        if isinstance(type_declaration, MutableSequence):

            sub_names: List[str] = list(
                dict.fromkeys([self.type_loader(i).name for i in type_declaration])
            )
            return self.declare_type(
                TypeDef(
                    "union_of_{}".format("_or_".join(sub_names)),
                    "_UnionLoader(({},))".format(", ".join(sub_names)),
                )
            )
        if isinstance(type_declaration, MutableMapping):
            if type_declaration["type"] in (
                "array",
                "https://w3id.org/cwl/salad#array",
            ):
                i = self.type_loader(type_declaration["items"])
                return self.declare_type(
                    TypeDef(f"array_of_{i.name}", f"_ArrayLoader({i.name})")
                )
            if type_declaration["type"] in ("enum", "https://w3id.org/cwl/salad#enum"):
                for sym in type_declaration["symbols"]:
                    self.add_vocab(shortname(sym), sym)
                return self.declare_type(
                    TypeDef(
                        self.safe_name(type_declaration["name"]) + "Loader",
                        '_EnumLoader(("{}",))'.format(
                            '", "'.join(
                                self.safe_name(sym)
                                for sym in type_declaration["symbols"]
                            )
                        ),
                    )
                )

            if type_declaration["type"] in (
                "record",
                "https://w3id.org/cwl/salad#record",
            ):
                return self.declare_type(
                    TypeDef(
                        self.safe_name(type_declaration["name"]) + "Loader",
                        "_RecordLoader({})".format(
                            self.safe_name(type_declaration["name"]),
                        ),
                        abstract=type_declaration.get("abstract", False),
                    )
                )
            raise SchemaException("wft {}".format(type_declaration["type"]))

        if type_declaration in prims:
            return prims[type_declaration]

        if type_declaration in ("Expression", "https://w3id.org/cwl/cwl#Expression"):
            return self.declare_type(
                TypeDef(
                    self.safe_name(type_declaration) + "Loader",
                    "_ExpressionLoader(str)",
                )
            )
        return self.collected_types[self.safe_name(type_declaration) + "Loader"]

    def declare_id_field(
        self,
        name: str,
        fieldtype: TypeDef,
        doc: str,
        optional: bool,
        subscope: Optional[str],
    ) -> None:
        if self.current_class_is_abstract:
            return

        self.declare_field(name, fieldtype, doc, True)

        if optional:
            opt = """{safename} = "_:" + str(_uuid__.uuid4())""".format(
                safename=self.safe_name(name)
            )
        else:
            opt = """raise ValidationException("Missing {fieldname}")""".format(
                fieldname=shortname(name)
            )

        if subscope is not None:
            name = name + subscope

        self.out.write(
            """
        __original_{safename}_is_none = {safename} is None
        if {safename} is None:
            if docRoot is not None:
                {safename} = docRoot
            else:
                {opt}
        if not __original_{safename}_is_none:
            baseuri = {safename}
""".format(
                safename=self.safe_name(name), opt=opt
            )
        )

    def declare_field(
        self, name: str, fieldtype: TypeDef, doc: Optional[str], optional: bool
    ) -> None:

        if self.current_class_is_abstract:
            return

        if shortname(name) == "class":
            return

        if optional:
            self.out.write(f"""        if "{shortname(name)}" in _doc:\n""")
            spc = "    "
        else:
            spc = ""
        self.out.write(
            """{spc}        try:
{spc}            {safename} = load_field(
{spc}                _doc.get("{fieldname}"),
{spc}                {fieldtype},
{spc}                baseuri,
{spc}                loadingOptions,
{spc}            )
{spc}        except ValidationException as e:
{spc}            _errors__.append(
{spc}                ValidationException(
{spc}                    \"the `{fieldname}` field is not valid because:\",
{spc}                    SourceLine(_doc, "{fieldname}", str),
{spc}                    [e],
{spc}                )
{spc}            )
""".format(
                safename=self.safe_name(name),
                fieldname=shortname(name),
                fieldtype=fieldtype.name,
                spc=spc,
            )
        )
        if optional:
            self.out.write(
                """        else:
            {safename} = None
""".format(
                    safename=self.safe_name(name)
                )
            )

        if name == self.idfield or not self.idfield:
            baseurl = "base_url"
        else:
            baseurl = f"self.{self.safe_name(self.idfield)}"

        if fieldtype.is_uri:
            self.serializer.write(
                fmt(
                    """
if self.{safename} is not None:
    u = save_relative_uri(self.{safename}, {baseurl}, {scoped_id}, {ref_scope}, relative_uris)
    if u:
        r["{fieldname}"] = u
""".format(
                        safename=self.safe_name(name),
                        fieldname=shortname(name).strip(),
                        baseurl=baseurl,
                        scoped_id=fieldtype.scoped_id,
                        ref_scope=fieldtype.ref_scope,
                    ),
                    8,
                )
            )
        else:
            self.serializer.write(
                fmt(
                    """
if self.{safename} is not None:
    r["{fieldname}"] = save(
        self.{safename}, top=False, base_url={baseurl}, relative_uris=relative_uris
    )
""".format(
                        safename=self.safe_name(name),
                        fieldname=shortname(name),
                        baseurl=baseurl,
                    ),
                    8,
                )
            )

    def uri_loader(
        self,
        inner: TypeDef,
        scoped_id: bool,
        vocab_term: bool,
        ref_scope: Optional[int],
    ) -> TypeDef:
        """Construct the TypeDef for the given URI loader."""
        return self.declare_type(
            TypeDef(
                f"uri_{inner.name}_{scoped_id}_{vocab_term}_{ref_scope}",
                "_URILoader({}, {}, {}, {})".format(
                    inner.name, scoped_id, vocab_term, ref_scope
                ),
                is_uri=True,
                scoped_id=scoped_id,
                ref_scope=ref_scope,
            )
        )

    def idmap_loader(
        self, field: str, inner: TypeDef, map_subject: str, map_predicate: Optional[str]
    ) -> TypeDef:
        """Construct the TypeDef for the given mapped ID loader."""
        return self.declare_type(
            TypeDef(
                f"idmap_{self.safe_name(field)}_{inner.name}",
                "_IdMapLoader({}, '{}', '{}')".format(
                    inner.name, map_subject, map_predicate
                ),
            )
        )

    def typedsl_loader(self, inner: TypeDef, ref_scope: Optional[int]) -> TypeDef:
        """Construct the TypeDef for the given DSL loader."""
        return self.declare_type(
            TypeDef(
                f"typedsl_{self.safe_name(inner.name)}_{ref_scope}",
                f"_TypeDSLLoader({self.safe_name(inner.name)}, {ref_scope})",
            )
        )

    def secondaryfilesdsl_loader(self, inner: TypeDef) -> TypeDef:
        """Construct the TypeDef for secondary files."""
        return self.declare_type(
            TypeDef(
                f"secondaryfilesdsl_{inner.name}",
                f"_SecondaryDSLLoader({inner.name})",
            )
        )

    def epilogue(self, root_loader: TypeDef) -> None:
        """Trigger to generate the epilouge code."""
        self.out.write("_vocab = {\n")
        for k in sorted(self.vocab.keys()):
            self.out.write(f'    "{k}": "{self.vocab[k]}",\n')
        self.out.write("}\n")

        self.out.write("_rvocab = {\n")
        for k in sorted(self.vocab.keys()):
            self.out.write(f'    "{self.vocab[k]}": "{k}",\n')
        self.out.write("}\n\n")

        for _, collected_type in self.collected_types.items():
            if not collected_type.abstract:
                self.out.write(
                    fmt(f"{collected_type.name} = {collected_type.init}\n", 0)
                )
        self.out.write("\n")

        self.out.write(
            """
def load_document(
    doc: Any,
    baseuri: Optional[str] = None,
    loadingOptions: Optional[LoadingOptions] = None,
) -> Any:
    if baseuri is None:
        baseuri = file_uri(os.getcwd()) + "/"
    if loadingOptions is None:
        loadingOptions = LoadingOptions()
    return _document_load(
        %(name)s,
        doc,
        baseuri,
        loadingOptions,
    )


def load_document_by_string(
    string: Any,
    uri: str,
    loadingOptions: Optional[LoadingOptions] = None,
) -> Any:
    yaml = yaml_no_ts()
    result = yaml.load(string)
    add_lc_filename(result, uri)

    if loadingOptions is None:
        loadingOptions = LoadingOptions(fileuri=uri)
    loadingOptions.idx[uri] = result

    return _document_load(
        %(name)s,
        result,
        uri,
        loadingOptions,
    )


def load_document_by_yaml(
    yaml: Any,
    uri: str,
    loadingOptions: Optional[LoadingOptions] = None,
) -> Any:
    """
            '"""'
            """
    Shortcut to load via a YAML object.
    yaml: must be from ruamel.yaml.main.YAML.load with preserve_quotes=True
    """
            '"""'
            """
    add_lc_filename(yaml, uri)

    if loadingOptions is None:
        loadingOptions = LoadingOptions(fileuri=uri)
    loadingOptions.idx[uri] = yaml

    return _document_load(
        %(name)s,
        yaml,
        uri,
        loadingOptions,
    )
"""
            % dict(name=root_loader.name)
        )



# Stubs for rdflib.plugins.parsers.notation3 (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

from rdflib.parser import Parser

class BadSyntax(SyntaxError):
    lines = ...  # type: Any
    def __init__(self, uri, lines, argstr, i, why): ...
    @property
    def message(self): ...



from typing import Any
from xml.sax import xmlreader

class Parser:
    def __init__(self) -> None: ...
    def parse(self, source, sink) -> None: ...

class InputSource(xmlreader.InputSource):
    content_type: Any
    auto_close: bool
    def __init__(self, system_id: Any | None = ...) -> None: ...
    def close(self) -> None: ...

class StringInputSource(InputSource):
    def __init__(self, value, system_id: Any | None = ...) -> None: ...

class URLInputSource(InputSource):
    url: Any
    content_type: Any
    response_info: Any
    def __init__(
        self, system_id: Any | None = ..., format: Any | None = ...
    ) -> None: ...

class FileInputSource(InputSource):
    file: Any
    def __init__(self, file) -> None: ...



# Stubs for cachecontrol.adapter (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

from requests.adapters import HTTPAdapter

from .cache import DictCache as DictCache
from .controller import CacheController as CacheController
from .filewrapper import CallbackFileWrapper as CallbackFileWrapper

class CacheControlAdapter(HTTPAdapter):
    invalidating_methods = ...  # type: Any
    cache = ...  # type: Any
    heuristic = ...  # type: Any
    controller = ...  # type: Any
    def __init__(
        self,
        cache=None,
        cache_etags=True,
        controller_class=None,
        serializer=None,
        heuristic=None,
        *args,
        **kw
    ): ...
    def send(self, request, **kw): ...
    def build_response(self, request, response, from_cache=False): ...
    def close(self): ...



# NOTE: This dynamically typed stub was automatically generated by stubgen.

from os import PathLike
from typing import Any, Callable, Union

from ..cache import BaseCache as BaseCache
from ..controller import CacheController as CacheController

class FileCache(BaseCache):
    directory = ...  # type: Union[str, PathLike[str]]
    forever = ...  # type: bool
    filemode = ...  # type: str
    dirmode = ...  # type: str
    lock_class = ...  # type: Callable
    def __init__(
        self,
        directory: Union[str, PathLike[str]],
        forever=False,
        filemode=384,
        dirmode=448,
        use_dir_lock=None,
        lock_class=None,
    ) -> None: ...
    @staticmethod
    def encode(x): ...
    def get(self, key): ...
    def set(self, key, value): ...
    def delete(self, key): ...

def url_to_file_path(url, filecache): ...



from typing import Any, Callable, Union

from .paths import AlternativePath, InvPath, NegatedPath, Path, SequencePath

class Node: ...

class Identifier(Node, str):
    def __new__(cls, value): ...
    def eq(self, other): ...
    def neq(self, other): ...
    def __ne__(self, other): ...
    def __eq__(self, other): ...
    def __gt__(self, other): ...
    def __lt__(self, other): ...
    def __le__(self, other): ...
    def __ge__(self, other): ...
    def startswith(self, prefix, start=..., end=...) -> bool: ...
    __hash__: Any

class URIRef(Identifier):
    __or__: Callable[[URIRef, Union[URIRef, Path]], AlternativePath]
    __invert__: Callable[[URIRef], InvPath]
    __neg__: Callable[[URIRef], NegatedPath]
    __truediv__: Callable[[URIRef, Union[URIRef, Path]], SequencePath]
    def __new__(cls, value, base: Any | None = ...): ...
    def toPython(self): ...
    def n3(self, namespace_manager: Any | None = ...) -> str: ...
    def defrag(self): ...
    def __reduce__(self): ...
    def __getnewargs__(self): ...
    def __add__(self, other): ...
    def __radd__(self, other): ...
    def __mod__(self, other): ...
    def de_skolemize(self): ...

class Genid(URIRef): ...
class RDFLibGenid(Genid): ...

class BNode(Identifier):
    def __new__(cls, value: Any | None = ..., _sn_gen=..., _prefix=...): ...
    def toPython(self): ...
    def n3(self, namespace_manager: Any | None = ...): ...
    def __getnewargs__(self): ...
    def __reduce__(self): ...
    def skolemize(self, authority: Any | None = ..., basepath: Any | None = ...): ...

class Literal(Identifier):
    __doc__: str
    def __new__(
        cls,
        lexical_or_value,
        lang: Any | None = ...,
        datatype: Any | None = ...,
        normalize: Any | None = ...,
    ): ...
    def normalize(self): ...
    @property
    def value(self): ...
    @property
    def language(self): ...
    @property
    def datatype(self): ...
    def __reduce__(self): ...
    def __add__(self, val): ...
    def __bool__(self): ...
    def __neg__(self): ...
    def __pos__(self): ...
    def __abs__(self): ...
    def __invert__(self): ...
    def __gt__(self, other): ...
    def __lt__(self, other): ...
    def __le__(self, other): ...
    def __ge__(self, other): ...
    def __hash__(self): ...
    def __eq__(self, other): ...
    def eq(self, other): ...
    def neq(self, other): ...
    def n3(self, namespace_manager: Any | None = ...): ...
    def toPython(self): ...

def bind(
    datatype,
    pythontype,
    constructor: Any | None = ...,
    lexicalizer: Any | None = ...,
    datatype_specific: bool = ...,
) -> None: ...

class Variable(Identifier):
    def __new__(cls, value): ...
    def toPython(self): ...
    def n3(self, namespace_manager: Any | None = ...): ...
    def __reduce__(self): ...

class Statement(Node, tuple):
    def __new__(cls, triple, context): ...
    def __reduce__(self): ...
    def toPython(self): ...



"""Resource fetching."""
import logging
import os
import re
import sys
import urllib
from typing import List, Optional

import requests

from .exceptions import ValidationException
from .utils import CacheType

_re_drive = re.compile(r"/([a-zA-Z]):")
_logger = logging.getLogger("salad")


class Fetcher:
    def __init__(
        self,
        cache: CacheType,
        session: Optional[requests.sessions.Session],
    ) -> None:
        pass

    def fetch_text(self, url: str, content_types: Optional[List[str]] = None) -> str:
        raise NotImplementedError()

    def check_exists(self, url: str) -> bool:
        raise NotImplementedError()

    def urljoin(self, base_url: str, url: str) -> str:
        raise NotImplementedError()

    schemes = ["file", "http", "https", "mailto"]

    def supported_schemes(self) -> List[str]:
        return self.schemes


class DefaultFetcher(Fetcher):
    def __init__(
        self,
        cache: CacheType,
        session: Optional[requests.sessions.Session],
    ) -> None:
        self.cache = cache
        self.session = session

    def fetch_text(self, url: str, content_types: Optional[List[str]] = None) -> str:
        """Retrieve the given resource as a string."""
        result = self.cache.get(url, None)
        if isinstance(result, str):
            return result

        split = urllib.parse.urlsplit(url)
        scheme, path = split.scheme, split.path

        if scheme in ["http", "https"] and self.session is not None:
            try:
                headers = {}
                if content_types:
                    headers["Accept"] = ", ".join(content_types) + ", */*;q=0.8"
                resp = self.session.get(url, headers=headers)
                resp.raise_for_status()
            except Exception as e:
                raise ValidationException(f"Error fetching {url}: {e}") from e
            if content_types and "content-type" in resp.headers:
                content_type = resp.headers["content-type"].split(";")[:1][0]
                if content_type not in content_types:
                    _logger.warning(
                        f"While fetching {url}, got content-type of "
                        f"'{content_type}'. Expected one of {content_types}."
                    )
            return resp.text
        if scheme == "file":
            try:
                # On Windows, url.path will be /drive:/path ; on Unix systems,
                # /path. As we want drive:/path instead of /drive:/path on Windows,
                # remove the leading /.
                if os.path.isabs(
                    path[1:]
                ):  # checking if pathis valid after removing front / or not
                    path = path[1:]
                with open(
                    urllib.request.url2pathname(str(path)), encoding="utf-8"
                ) as fp:
                    return str(fp.read())

            except OSError as err:
                if err.filename == path:
                    raise ValidationException(str(err)) from err
                else:
                    raise ValidationException(f"Error reading {url}: {err}") from err
        raise ValidationException(f"Unsupported scheme in url: {url}")

    def check_exists(self, url: str) -> bool:
        if url in self.cache:
            return True

        split = urllib.parse.urlsplit(url)
        scheme, path = split.scheme, split.path

        if scheme in ["http", "https"]:
            if self.session is None:
                raise ValidationException(f"Can't check {scheme} URL, session is None")
            try:
                resp = self.session.head(url, allow_redirects=True)
                resp.raise_for_status()
            except Exception:
                return False
            self.cache[url] = True
            return True
        if scheme == "file":
            return os.path.exists(urllib.request.url2pathname(str(path)))
        if scheme == "mailto":
            return True
        raise ValidationException(f"Unsupported scheme '{scheme}' in url: {url}")

    def urljoin(self, base_url: str, url: str) -> str:
        if url.startswith("_:"):
            return url

        basesplit = urllib.parse.urlsplit(base_url)
        split = urllib.parse.urlsplit(url)
        if basesplit.scheme and basesplit.scheme != "file" and split.scheme == "file":
            raise ValidationException(
                "Not resolving potential remote exploit {} from base {}".format(
                    url, base_url
                )
            )

        if sys.platform == "win32":
            if base_url == url:
                return url
            basesplit = urllib.parse.urlsplit(base_url)
            # note that below might split
            # "C:" with "C" as URI scheme
            split = urllib.parse.urlsplit(url)

            has_drive = split.scheme and len(split.scheme) == 1

            if basesplit.scheme == "file":
                # Special handling of relative file references on Windows
                # as urllib seems to not be quite up to the job

                # netloc MIGHT appear in equivalents of UNC Strings
                # \\server1.example.com\path as
                # file:///server1.example.com/path
                # https://tools.ietf.org/html/rfc8089#appendix-E.3.2
                # (TODO: test this)
                netloc = split.netloc or basesplit.netloc

                # Check if url is a local path like "C:/Users/fred"
                # or actually an absolute URI like http://example.com/fred
                if has_drive:
                    # Assume split.scheme is actually a drive, e.g. "C:"
                    # so we'll recombine into a path
                    path_with_drive = urllib.parse.urlunsplit(
                        (split.scheme, "", split.path, "", "")
                    )
                    # Compose new file:/// URI with path_with_drive
                    # .. carrying over any #fragment (?query just in case..)
                    return urllib.parse.urlunsplit(
                        ("file", netloc, path_with_drive, split.query, split.fragment)
                    )
                if (
                    not split.scheme
                    and not netloc
                    and split.path
                    and split.path.startswith("/")
                ):
                    # Relative - but does it have a drive?
                    base_drive = _re_drive.match(basesplit.path)
                    drive = _re_drive.match(split.path)
                    if base_drive and not drive:
                        # Keep drive letter from base_url
                        # https://tools.ietf.org/html/rfc8089#appendix-E.2.1
                        # e.g. urljoin("file:///D:/bar/a.txt", "/foo/b.txt")
                        #          == file:///D:/foo/b.txt
                        path_with_drive = "/{}:{}".format(
                            base_drive.group(1), split.path
                        )
                        return urllib.parse.urlunsplit(
                            (
                                "file",
                                netloc,
                                path_with_drive,
                                split.query,
                                split.fragment,
                            )
                        )

                # else: fall-through to resolve as relative URI
            elif has_drive:
                # Base is http://something but url is C:/something - which urllib
                # would wrongly resolve as an absolute path that could later be used
                # to access local files
                raise ValidationException(
                    "Not resolving potential remote exploit {} from base {}".format(
                        url, base_url
                    )
                )

        return urllib.parse.urljoin(base_url, url)



from typing import Any

class Error(Exception):
    msg: Any
    def __init__(self, msg: Any | None = ...) -> None: ...

class TypeCheckError(Error):
    type: Any
    node: Any
    def __init__(self, node) -> None: ...

class SubjectTypeError(TypeCheckError):
    msg: Any
    def __init__(self, node) -> None: ...

class PredicateTypeError(TypeCheckError):
    msg: Any
    def __init__(self, node) -> None: ...

class ObjectTypeError(TypeCheckError):
    msg: Any
    def __init__(self, node) -> None: ...

class ContextTypeError(TypeCheckError):
    msg: Any
    def __init__(self, node) -> None: ...

class ParserError(Error):
    msg: Any
    def __init__(self, msg) -> None: ...

class UniquenessError(Error):
    def __init__(self, values) -> None: ...



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class SDO(DefinedNamespace):
    ActiveActionStatus: URIRef
    CompletedActionStatus: URIRef
    FailedActionStatus: URIRef
    PotentialActionStatus: URIRef
    Researcher: URIRef
    GroupBoardingPolicy: URIRef
    ZoneBoardingPolicy: URIRef
    AudiobookFormat: URIRef
    EBook: URIRef
    Hardcover: URIRef
    Paperback: URIRef
    HearingImpairedSupported: URIRef
    TollFree: URIRef
    Friday: URIRef
    Monday: URIRef
    PublicHolidays: URIRef
    Saturday: URIRef
    Sunday: URIRef
    Thursday: URIRef
    Tuesday: URIRef
    Wednesday: URIRef
    OnSitePickup: URIRef
    CommentPermission: URIRef
    ReadPermission: URIRef
    WritePermission: URIRef
    AllWheelDriveConfiguration: URIRef
    FourWheelDriveConfiguration: URIRef
    FrontWheelDriveConfiguration: URIRef
    RearWheelDriveConfiguration: URIRef
    EventCancelled: URIRef
    EventMovedOnline: URIRef
    EventPostponed: URIRef
    EventRescheduled: URIRef
    EventScheduled: URIRef
    CoOp: URIRef
    MultiPlayer: URIRef
    SinglePlayer: URIRef
    OfflinePermanently: URIRef
    OfflineTemporarily: URIRef
    Online: URIRef
    OnlineFull: URIRef
    Female: URIRef
    Male: URIRef
    Discontinued: URIRef
    InStock: URIRef
    InStoreOnly: URIRef
    LimitedAvailability: URIRef
    OnlineOnly: URIRef
    OutOfStock: URIRef
    PreOrder: URIRef
    PreSale: URIRef
    SoldOut: URIRef
    ItemListOrderAscending: URIRef
    ItemListOrderDescending: URIRef
    ItemListUnordered: URIRef
    ParkingMap: URIRef
    SeatingMap: URIRef
    TransitMap: URIRef
    VenueMap: URIRef
    CompilationAlbum: URIRef
    DJMixAlbum: URIRef
    DemoAlbum: URIRef
    LiveAlbum: URIRef
    MixtapeAlbum: URIRef
    RemixAlbum: URIRef
    SoundtrackAlbum: URIRef
    SpokenWordAlbum: URIRef
    StudioAlbum: URIRef
    AlbumRelease: URIRef
    BroadcastRelease: URIRef
    EPRelease: URIRef
    SingleRelease: URIRef
    CDFormat: URIRef
    CassetteFormat: URIRef
    DVDFormat: URIRef
    DigitalAudioTapeFormat: URIRef
    DigitalFormat: URIRef
    LaserDiscFormat: URIRef
    VinylFormat: URIRef
    DamagedCondition: URIRef
    NewCondition: URIRef
    RefurbishedCondition: URIRef
    UsedCondition: URIRef
    OrderCancelled: URIRef
    OrderDelivered: URIRef
    OrderInTransit: URIRef
    OrderPaymentDue: URIRef
    OrderPickupAvailable: URIRef
    OrderProblem: URIRef
    OrderProcessing: URIRef
    OrderReturned: URIRef
    PaymentAutomaticallyApplied: URIRef
    PaymentComplete: URIRef
    PaymentDeclined: URIRef
    PaymentDue: URIRef
    PaymentPastDue: URIRef
    ReservationCancelled: URIRef
    ReservationConfirmed: URIRef
    ReservationHold: URIRef
    ReservationPending: URIRef
    DiabeticDiet: URIRef
    GlutenFreeDiet: URIRef
    HalalDiet: URIRef
    HinduDiet: URIRef
    KosherDiet: URIRef
    LowCalorieDiet: URIRef
    LowFatDiet: URIRef
    LowLactoseDiet: URIRef
    LowSaltDiet: URIRef
    VeganDiet: URIRef
    VegetarianDiet: URIRef
    RsvpResponseMaybe: URIRef
    RsvpResponseNo: URIRef
    RsvpResponseYes: URIRef
    LeftHandDriving: URIRef
    RightHandDriving: URIRef
    about: URIRef
    acceptedAnswer: URIRef
    acceptedOffer: URIRef
    acceptedPaymentMethod: URIRef
    acceptsReservations: URIRef
    accessCode: URIRef
    accessMode: URIRef
    accessModeSufficient: URIRef
    accessibilityAPI: URIRef
    accessibilityControl: URIRef
    accessibilityFeature: URIRef
    accessibilityHazard: URIRef
    accessibilitySummary: URIRef
    accountId: URIRef
    accountablePerson: URIRef
    acquiredFrom: URIRef
    actionAccessibilityRequirement: URIRef
    actionApplication: URIRef
    actionOption: URIRef
    actionPlatform: URIRef
    actionStatus: URIRef
    actor: URIRef
    actors: URIRef
    addOn: URIRef
    additionalName: URIRef
    additionalNumberOfGuests: URIRef
    additionalProperty: URIRef
    additionalType: URIRef
    address: URIRef
    addressCountry: URIRef
    addressLocality: URIRef
    addressRegion: URIRef
    advanceBookingRequirement: URIRef
    affiliation: URIRef
    afterMedia: URIRef
    agent: URIRef
    aggregateRating: URIRef
    aircraft: URIRef
    album: URIRef
    albumProductionType: URIRef
    albumRelease: URIRef
    albumReleaseType: URIRef
    albums: URIRef
    alignmentType: URIRef
    alternateName: URIRef
    alternativeHeadline: URIRef
    alumni: URIRef
    alumniOf: URIRef
    amenityFeature: URIRef
    amount: URIRef
    amountOfThisGood: URIRef
    annualPercentageRate: URIRef
    answerCount: URIRef
    application: URIRef
    applicationCategory: URIRef
    applicationSubCategory: URIRef
    applicationSuite: URIRef
    appliesToDeliveryMethod: URIRef
    appliesToPaymentMethod: URIRef
    area: URIRef
    areaServed: URIRef
    arrivalAirport: URIRef
    arrivalBusStop: URIRef
    arrivalGate: URIRef
    arrivalPlatform: URIRef
    arrivalStation: URIRef
    arrivalTerminal: URIRef
    arrivalTime: URIRef
    artEdition: URIRef
    artMedium: URIRef
    artform: URIRef
    articleBody: URIRef
    articleSection: URIRef
    artworkSurface: URIRef
    assembly: URIRef
    assemblyVersion: URIRef
    associatedArticle: URIRef
    associatedMedia: URIRef
    athlete: URIRef
    attendee: URIRef
    attendees: URIRef
    audience: URIRef
    audienceType: URIRef
    audio: URIRef
    authenticator: URIRef
    author: URIRef
    availability: URIRef
    availabilityEnds: URIRef
    availabilityStarts: URIRef
    availableAtOrFrom: URIRef
    availableChannel: URIRef
    availableDeliveryMethod: URIRef
    availableFrom: URIRef
    availableLanguage: URIRef
    availableOnDevice: URIRef
    availableThrough: URIRef
    award: URIRef
    awards: URIRef
    awayTeam: URIRef
    baseSalary: URIRef
    bccRecipient: URIRef
    bed: URIRef
    beforeMedia: URIRef
    benefits: URIRef
    bestRating: URIRef
    billingAddress: URIRef
    billingIncrement: URIRef
    billingPeriod: URIRef
    birthDate: URIRef
    birthPlace: URIRef
    bitrate: URIRef
    blogPost: URIRef
    blogPosts: URIRef
    boardingGroup: URIRef
    boardingPolicy: URIRef
    bookEdition: URIRef
    bookFormat: URIRef
    bookingAgent: URIRef
    bookingTime: URIRef
    borrower: URIRef
    box: URIRef
    branchCode: URIRef
    branchOf: URIRef
    brand: URIRef
    breadcrumb: URIRef
    broadcastAffiliateOf: URIRef
    broadcastChannelId: URIRef
    broadcastDisplayName: URIRef
    broadcastFrequency: URIRef
    broadcastFrequencyValue: URIRef
    broadcastOfEvent: URIRef
    broadcastServiceTier: URIRef
    broadcastTimezone: URIRef
    broadcaster: URIRef
    broker: URIRef
    browserRequirements: URIRef
    busName: URIRef
    busNumber: URIRef
    businessFunction: URIRef
    buyer: URIRef
    byArtist: URIRef
    calories: URIRef
    candidate: URIRef
    caption: URIRef
    carbohydrateContent: URIRef
    cargoVolume: URIRef
    carrier: URIRef
    carrierRequirements: URIRef
    catalog: URIRef
    catalogNumber: URIRef
    category: URIRef
    ccRecipient: URIRef
    character: URIRef
    characterAttribute: URIRef
    characterName: URIRef
    cheatCode: URIRef
    checkinTime: URIRef
    checkoutTime: URIRef
    childMaxAge: URIRef
    childMinAge: URIRef
    children: URIRef
    cholesterolContent: URIRef
    circle: URIRef
    citation: URIRef
    claimReviewed: URIRef
    clipNumber: URIRef
    closes: URIRef
    coach: URIRef
    codeRepository: URIRef
    codeSampleType: URIRef
    colleague: URIRef
    colleagues: URIRef
    collection: URIRef
    color: URIRef
    comment: URIRef
    commentCount: URIRef
    commentText: URIRef
    commentTime: URIRef
    competitor: URIRef
    composer: URIRef
    confirmationNumber: URIRef
    contactOption: URIRef
    contactPoint: URIRef
    contactPoints: URIRef
    contactType: URIRef
    containedIn: URIRef
    containedInPlace: URIRef
    containsPlace: URIRef
    containsSeason: URIRef
    contentLocation: URIRef
    contentRating: URIRef
    contentSize: URIRef
    contentType: URIRef
    contentUrl: URIRef
    contributor: URIRef
    cookTime: URIRef
    cookingMethod: URIRef
    copyrightHolder: URIRef
    copyrightYear: URIRef
    countriesNotSupported: URIRef
    countriesSupported: URIRef
    countryOfOrigin: URIRef
    course: URIRef
    courseCode: URIRef
    courseMode: URIRef
    coursePrerequisites: URIRef
    coverageEndTime: URIRef
    coverageStartTime: URIRef
    creator: URIRef
    creditedTo: URIRef
    cssSelector: URIRef
    currenciesAccepted: URIRef
    currency: URIRef
    customer: URIRef
    dataFeedElement: URIRef
    dataset: URIRef
    datasetTimeInterval: URIRef
    dateCreated: URIRef
    dateDeleted: URIRef
    dateIssued: URIRef
    dateModified: URIRef
    datePosted: URIRef
    datePublished: URIRef
    dateRead: URIRef
    dateReceived: URIRef
    dateSent: URIRef
    dateVehicleFirstRegistered: URIRef
    dateline: URIRef
    dayOfWeek: URIRef
    deathDate: URIRef
    deathPlace: URIRef
    defaultValue: URIRef
    deliveryAddress: URIRef
    deliveryLeadTime: URIRef
    deliveryMethod: URIRef
    deliveryStatus: URIRef
    department: URIRef
    departureAirport: URIRef
    departureBusStop: URIRef
    departureGate: URIRef
    departurePlatform: URIRef
    departureStation: URIRef
    departureTerminal: URIRef
    departureTime: URIRef
    dependencies: URIRef
    depth: URIRef
    description: URIRef
    device: URIRef
    director: URIRef
    directors: URIRef
    disambiguatingDescription: URIRef
    discount: URIRef
    discountCode: URIRef
    discountCurrency: URIRef
    discusses: URIRef
    discussionUrl: URIRef
    dissolutionDate: URIRef
    distance: URIRef
    distribution: URIRef
    doorTime: URIRef
    downloadUrl: URIRef
    downvoteCount: URIRef
    driveWheelConfiguration: URIRef
    dropoffLocation: URIRef
    dropoffTime: URIRef
    duns: URIRef
    duration: URIRef
    durationOfWarranty: URIRef
    duringMedia: URIRef
    editor: URIRef
    educationalAlignment: URIRef
    educationalCredentialAwarded: URIRef
    educationalFramework: URIRef
    educationalRole: URIRef
    educationalUse: URIRef
    elevation: URIRef
    eligibleCustomerType: URIRef
    eligibleDuration: URIRef
    eligibleQuantity: URIRef
    eligibleRegion: URIRef
    eligibleTransactionVolume: URIRef
    email: URIRef
    embedUrl: URIRef
    employee: URIRef
    employees: URIRef
    employmentType: URIRef
    encodesCreativeWork: URIRef
    encoding: URIRef
    encodingFormat: URIRef
    encodingType: URIRef
    encodings: URIRef
    endDate: URIRef
    endTime: URIRef
    endorsee: URIRef
    entertainmentBusiness: URIRef
    episode: URIRef
    episodeNumber: URIRef
    episodes: URIRef
    equal: URIRef
    error: URIRef
    estimatedCost: URIRef
    estimatedFlightDuration: URIRef
    estimatedSalary: URIRef
    event: URIRef
    eventStatus: URIRef
    events: URIRef
    exampleOfWork: URIRef
    executableLibraryName: URIRef
    exerciseCourse: URIRef
    exifData: URIRef
    expectedArrivalFrom: URIRef
    expectedArrivalUntil: URIRef
    expectsAcceptanceOf: URIRef
    experienceRequirements: URIRef
    expires: URIRef
    familyName: URIRef
    fatContent: URIRef
    faxNumber: URIRef
    featureList: URIRef
    feesAndCommissionsSpecification: URIRef
    fiberContent: URIRef
    fileFormat: URIRef
    fileSize: URIRef
    firstPerformance: URIRef
    flightDistance: URIRef
    flightNumber: URIRef
    floorSize: URIRef
    followee: URIRef
    follows: URIRef
    foodEstablishment: URIRef
    foodEvent: URIRef
    founder: URIRef
    founders: URIRef
    foundingDate: URIRef
    foundingLocation: URIRef
    free: URIRef
    fromLocation: URIRef
    fuelConsumption: URIRef
    fuelEfficiency: URIRef
    fuelType: URIRef
    funder: URIRef
    game: URIRef
    gameItem: URIRef
    gameLocation: URIRef
    gamePlatform: URIRef
    gameServer: URIRef
    gameTip: URIRef
    genre: URIRef
    geo: URIRef
    geoContains: URIRef
    geoCoveredBy: URIRef
    geoCovers: URIRef
    geoCrosses: URIRef
    geoDisjoint: URIRef
    geoEquals: URIRef
    geoIntersects: URIRef
    geoMidpoint: URIRef
    geoOverlaps: URIRef
    geoRadius: URIRef
    geoTouches: URIRef
    geoWithin: URIRef
    geographicArea: URIRef
    givenName: URIRef
    globalLocationNumber: URIRef
    grantee: URIRef
    greater: URIRef
    greaterOrEqual: URIRef
    gtin12: URIRef
    gtin13: URIRef
    gtin14: URIRef
    gtin8: URIRef
    hasBroadcastChannel: URIRef
    hasCourseInstance: URIRef
    hasDeliveryMethod: URIRef
    hasDigitalDocumentPermission: URIRef
    hasMap: URIRef
    hasMenu: URIRef
    hasMenuItem: URIRef
    hasMenuSection: URIRef
    hasOccupation: URIRef
    hasOfferCatalog: URIRef
    hasPOS: URIRef
    hasPart: URIRef
    headline: URIRef
    height: URIRef
    highPrice: URIRef
    hiringOrganization: URIRef
    homeLocation: URIRef
    homeTeam: URIRef
    honorificPrefix: URIRef
    honorificSuffix: URIRef
    hostingOrganization: URIRef
    hoursAvailable: URIRef
    httpMethod: URIRef
    iataCode: URIRef
    icaoCode: URIRef
    identifier: URIRef
    illustrator: URIRef
    image: URIRef
    inAlbum: URIRef
    inBroadcastLineup: URIRef
    inLanguage: URIRef
    inPlaylist: URIRef
    incentiveCompensation: URIRef
    incentives: URIRef
    includedComposition: URIRef
    includedDataCatalog: URIRef
    includedInDataCatalog: URIRef
    includesObject: URIRef
    industry: URIRef
    ingredients: URIRef
    installUrl: URIRef
    instructor: URIRef
    instrument: URIRef
    interactionCount: URIRef
    interactionService: URIRef
    interactionStatistic: URIRef
    interactionType: URIRef
    interactivityType: URIRef
    interestRate: URIRef
    inventoryLevel: URIRef
    isAccessibleForFree: URIRef
    isAccessoryOrSparePartFor: URIRef
    isBasedOn: URIRef
    isBasedOnUrl: URIRef
    isConsumableFor: URIRef
    isFamilyFriendly: URIRef
    isGift: URIRef
    isLiveBroadcast: URIRef
    isPartOf: URIRef
    isRelatedTo: URIRef
    isSimilarTo: URIRef
    isVariantOf: URIRef
    isbn: URIRef
    isicV4: URIRef
    isrcCode: URIRef
    issn: URIRef
    issueNumber: URIRef
    issuedBy: URIRef
    issuedThrough: URIRef
    iswcCode: URIRef
    item: URIRef
    itemCondition: URIRef
    itemListElement: URIRef
    itemListOrder: URIRef
    itemOffered: URIRef
    itemReviewed: URIRef
    itemShipped: URIRef
    jobBenefits: URIRef
    jobLocation: URIRef
    keywords: URIRef
    knownVehicleDamages: URIRef
    knows: URIRef
    landlord: URIRef
    language: URIRef
    lastReviewed: URIRef
    latitude: URIRef
    learningResourceType: URIRef
    legalName: URIRef
    leiCode: URIRef
    lender: URIRef
    lesser: URIRef
    lesserOrEqual: URIRef
    license: URIRef
    line: URIRef
    liveBlogUpdate: URIRef
    loanTerm: URIRef
    location: URIRef
    locationCreated: URIRef
    lodgingUnitDescription: URIRef
    lodgingUnitType: URIRef
    logo: URIRef
    longitude: URIRef
    loser: URIRef
    lowPrice: URIRef
    lyricist: URIRef
    lyrics: URIRef
    mainContentOfPage: URIRef
    mainEntity: URIRef
    mainEntityOfPage: URIRef
    makesOffer: URIRef
    manufacturer: URIRef
    map: URIRef
    mapType: URIRef
    maps: URIRef
    material: URIRef
    maxPrice: URIRef
    maxValue: URIRef
    maximumAttendeeCapacity: URIRef
    mealService: URIRef
    median: URIRef
    member: URIRef
    memberOf: URIRef
    members: URIRef
    membershipNumber: URIRef
    memoryRequirements: URIRef
    mentions: URIRef
    menu: URIRef
    menuAddOn: URIRef
    merchant: URIRef
    messageAttachment: URIRef
    mileageFromOdometer: URIRef
    minPrice: URIRef
    minValue: URIRef
    minimumPaymentDue: URIRef
    model: URIRef
    modifiedTime: URIRef
    mpn: URIRef
    multipleValues: URIRef
    musicArrangement: URIRef
    musicBy: URIRef
    musicCompositionForm: URIRef
    musicGroupMember: URIRef
    musicReleaseFormat: URIRef
    musicalKey: URIRef
    naics: URIRef
    name: URIRef
    namedPosition: URIRef
    nationality: URIRef
    netWorth: URIRef
    nextItem: URIRef
    nonEqual: URIRef
    numAdults: URIRef
    numChildren: URIRef
    numTracks: URIRef
    numberOfAirbags: URIRef
    numberOfAxles: URIRef
    numberOfBeds: URIRef
    numberOfDoors: URIRef
    numberOfEmployees: URIRef
    numberOfEpisodes: URIRef
    numberOfForwardGears: URIRef
    numberOfItems: URIRef
    numberOfPages: URIRef
    numberOfPlayers: URIRef
    numberOfPreviousOwners: URIRef
    numberOfRooms: URIRef
    numberOfSeasons: URIRef
    numberedPosition: URIRef
    nutrition: URIRef
    object: URIRef
    occupancy: URIRef
    occupationLocation: URIRef
    offerCount: URIRef
    offeredBy: URIRef
    offers: URIRef
    openingHours: URIRef
    openingHoursSpecification: URIRef
    opens: URIRef
    operatingSystem: URIRef
    opponent: URIRef
    option: URIRef
    orderDate: URIRef
    orderDelivery: URIRef
    orderItemNumber: URIRef
    orderItemStatus: URIRef
    orderNumber: URIRef
    orderQuantity: URIRef
    orderStatus: URIRef
    orderedItem: URIRef
    organizer: URIRef
    originAddress: URIRef
    ownedFrom: URIRef
    ownedThrough: URIRef
    owns: URIRef
    pageEnd: URIRef
    pageStart: URIRef
    pagination: URIRef
    parent: URIRef
    parentItem: URIRef
    parentOrganization: URIRef
    parentService: URIRef
    parents: URIRef
    partOfEpisode: URIRef
    partOfInvoice: URIRef
    partOfOrder: URIRef
    partOfSeason: URIRef
    partOfSeries: URIRef
    partOfTVSeries: URIRef
    participant: URIRef
    partySize: URIRef
    passengerPriorityStatus: URIRef
    passengerSequenceNumber: URIRef
    paymentAccepted: URIRef
    paymentDue: URIRef
    paymentDueDate: URIRef
    paymentMethod: URIRef
    paymentMethodId: URIRef
    paymentStatus: URIRef
    paymentUrl: URIRef
    percentile10: URIRef
    percentile25: URIRef
    percentile75: URIRef
    percentile90: URIRef
    performTime: URIRef
    performer: URIRef
    performerIn: URIRef
    performers: URIRef
    permissionType: URIRef
    permissions: URIRef
    permitAudience: URIRef
    permittedUsage: URIRef
    petsAllowed: URIRef
    photo: URIRef
    photos: URIRef
    pickupLocation: URIRef
    pickupTime: URIRef
    playMode: URIRef
    playerType: URIRef
    playersOnline: URIRef
    polygon: URIRef
    position: URIRef
    postOfficeBoxNumber: URIRef
    postalCode: URIRef
    potentialAction: URIRef
    predecessorOf: URIRef
    prepTime: URIRef
    previousItem: URIRef
    previousStartDate: URIRef
    price: URIRef
    priceComponent: URIRef
    priceCurrency: URIRef
    priceRange: URIRef
    priceSpecification: URIRef
    priceType: URIRef
    priceValidUntil: URIRef
    primaryImageOfPage: URIRef
    printColumn: URIRef
    printEdition: URIRef
    printPage: URIRef
    printSection: URIRef
    processingTime: URIRef
    processorRequirements: URIRef
    producer: URIRef
    produces: URIRef
    productID: URIRef
    productSupported: URIRef
    productionCompany: URIRef
    productionDate: URIRef
    proficiencyLevel: URIRef
    programMembershipUsed: URIRef
    programName: URIRef
    programmingLanguage: URIRef
    programmingModel: URIRef
    propertyID: URIRef
    proteinContent: URIRef
    provider: URIRef
    providerMobility: URIRef
    providesBroadcastService: URIRef
    providesService: URIRef
    publicAccess: URIRef
    publication: URIRef
    publishedOn: URIRef
    publisher: URIRef
    publishingPrinciples: URIRef
    purchaseDate: URIRef
    query: URIRef
    quest: URIRef
    question: URIRef
    ratingCount: URIRef
    ratingValue: URIRef
    readonlyValue: URIRef
    realEstateAgent: URIRef
    recipe: URIRef
    recipeCategory: URIRef
    recipeCuisine: URIRef
    recipeIngredient: URIRef
    recipeInstructions: URIRef
    recipeYield: URIRef
    recipient: URIRef
    recordLabel: URIRef
    recordedAs: URIRef
    recordedAt: URIRef
    recordedIn: URIRef
    recordingOf: URIRef
    referenceQuantity: URIRef
    referencesOrder: URIRef
    regionsAllowed: URIRef
    relatedLink: URIRef
    relatedTo: URIRef
    releaseDate: URIRef
    releaseNotes: URIRef
    releaseOf: URIRef
    releasedEvent: URIRef
    relevantOccupation: URIRef
    remainingAttendeeCapacity: URIRef
    replacee: URIRef
    replacer: URIRef
    replyToUrl: URIRef
    reportNumber: URIRef
    representativeOfPage: URIRef
    requiredCollateral: URIRef
    requiredGender: URIRef
    requiredMaxAge: URIRef
    requiredMinAge: URIRef
    requiredQuantity: URIRef
    requirements: URIRef
    requiresSubscription: URIRef
    reservationFor: URIRef
    reservationId: URIRef
    reservationStatus: URIRef
    reservedTicket: URIRef
    responsibilities: URIRef
    result: URIRef
    resultComment: URIRef
    resultReview: URIRef
    review: URIRef
    reviewAspect: URIRef
    reviewBody: URIRef
    reviewCount: URIRef
    reviewRating: URIRef
    reviewedBy: URIRef
    reviews: URIRef
    roleName: URIRef
    rsvpResponse: URIRef
    runtime: URIRef
    runtimePlatform: URIRef
    salaryCurrency: URIRef
    sameAs: URIRef
    sampleType: URIRef
    saturatedFatContent: URIRef
    scheduledPaymentDate: URIRef
    scheduledTime: URIRef
    schemaVersion: URIRef
    screenCount: URIRef
    screenshot: URIRef
    season: URIRef
    seasonNumber: URIRef
    seasons: URIRef
    seatNumber: URIRef
    seatRow: URIRef
    seatSection: URIRef
    seatingType: URIRef
    securityScreening: URIRef
    seeks: URIRef
    seller: URIRef
    sender: URIRef
    serialNumber: URIRef
    serverStatus: URIRef
    servesCuisine: URIRef
    serviceArea: URIRef
    serviceAudience: URIRef
    serviceLocation: URIRef
    serviceOperator: URIRef
    serviceOutput: URIRef
    servicePhone: URIRef
    servicePostalAddress: URIRef
    serviceSmsNumber: URIRef
    serviceType: URIRef
    serviceUrl: URIRef
    servingSize: URIRef
    sharedContent: URIRef
    sibling: URIRef
    siblings: URIRef
    significantLink: URIRef
    significantLinks: URIRef
    skills: URIRef
    sku: URIRef
    slogan: URIRef
    smokingAllowed: URIRef
    sodiumContent: URIRef
    softwareAddOn: URIRef
    softwareHelp: URIRef
    softwareRequirements: URIRef
    softwareVersion: URIRef
    sourceOrganization: URIRef
    spatial: URIRef
    spatialCoverage: URIRef
    speakable: URIRef
    specialCommitments: URIRef
    specialOpeningHoursSpecification: URIRef
    specialty: URIRef
    sponsor: URIRef
    sportsActivityLocation: URIRef
    sportsEvent: URIRef
    sportsTeam: URIRef
    spouse: URIRef
    starRating: URIRef
    startDate: URIRef
    startTime: URIRef
    steeringPosition: URIRef
    step: URIRef
    stepValue: URIRef
    steps: URIRef
    storageRequirements: URIRef
    streetAddress: URIRef
    subEvent: URIRef
    subEvents: URIRef
    subOrganization: URIRef
    subReservation: URIRef
    subjectOf: URIRef
    successorOf: URIRef
    sugarContent: URIRef
    suggestedAnswer: URIRef
    suggestedGender: URIRef
    suggestedMaxAge: URIRef
    suggestedMinAge: URIRef
    suitableForDiet: URIRef
    superEvent: URIRef
    supply: URIRef
    supportingData: URIRef
    surface: URIRef
    target: URIRef
    targetCollection: URIRef
    targetDescription: URIRef
    targetName: URIRef
    targetPlatform: URIRef
    targetProduct: URIRef
    targetUrl: URIRef
    taxID: URIRef
    telephone: URIRef
    temporal: URIRef
    temporalCoverage: URIRef
    text: URIRef
    thumbnail: URIRef
    thumbnailUrl: URIRef
    tickerSymbol: URIRef
    ticketNumber: URIRef
    ticketToken: URIRef
    ticketedSeat: URIRef
    timeRequired: URIRef
    title: URIRef
    toLocation: URIRef
    toRecipient: URIRef
    tool: URIRef
    totalPaymentDue: URIRef
    totalPrice: URIRef
    totalTime: URIRef
    touristType: URIRef
    track: URIRef
    trackingNumber: URIRef
    trackingUrl: URIRef
    tracks: URIRef
    trailer: URIRef
    trainName: URIRef
    trainNumber: URIRef
    transFatContent: URIRef
    transcript: URIRef
    translator: URIRef
    typeOfBed: URIRef
    typeOfGood: URIRef
    typicalAgeRange: URIRef
    underName: URIRef
    unitCode: URIRef
    unitText: URIRef
    unsaturatedFatContent: URIRef
    uploadDate: URIRef
    upvoteCount: URIRef
    url: URIRef
    urlTemplate: URIRef
    userInteractionCount: URIRef
    validFor: URIRef
    validFrom: URIRef
    validIn: URIRef
    validThrough: URIRef
    validUntil: URIRef
    value: URIRef
    valueAddedTaxIncluded: URIRef
    valueMaxLength: URIRef
    valueMinLength: URIRef
    valueName: URIRef
    valuePattern: URIRef
    valueReference: URIRef
    valueRequired: URIRef
    vatID: URIRef
    vehicleConfiguration: URIRef
    vehicleEngine: URIRef
    vehicleIdentificationNumber: URIRef
    vehicleInteriorColor: URIRef
    vehicleInteriorType: URIRef
    vehicleModelDate: URIRef
    vehicleSeatingCapacity: URIRef
    vehicleTransmission: URIRef
    vendor: URIRef
    version: URIRef
    video: URIRef
    videoFormat: URIRef
    videoFrameSize: URIRef
    videoQuality: URIRef
    volumeNumber: URIRef
    warranty: URIRef
    warrantyPromise: URIRef
    warrantyScope: URIRef
    webCheckinTime: URIRef
    weight: URIRef
    width: URIRef
    winner: URIRef
    wordCount: URIRef
    workExample: URIRef
    workFeatured: URIRef
    workHours: URIRef
    workLocation: URIRef
    workPerformed: URIRef
    workPresented: URIRef
    worksFor: URIRef
    worstRating: URIRef
    xpath: URIRef
    yearlyRevenue: URIRef
    yearsInOperation: URIRef
    AMRadioChannel: URIRef
    APIReference: URIRef
    AboutPage: URIRef
    AcceptAction: URIRef
    Accommodation: URIRef
    AccountingService: URIRef
    AchieveAction: URIRef
    Action: URIRef
    ActionAccessSpecification: URIRef
    ActionStatusType: URIRef
    ActivateAction: URIRef
    AddAction: URIRef
    AdministrativeArea: URIRef
    AdultEntertainment: URIRef
    AggregateOffer: URIRef
    AggregateRating: URIRef
    AgreeAction: URIRef
    Airline: URIRef
    Airport: URIRef
    AlignmentObject: URIRef
    AllocateAction: URIRef
    AmusementPark: URIRef
    AnimalShelter: URIRef
    Answer: URIRef
    Apartment: URIRef
    ApartmentComplex: URIRef
    AppendAction: URIRef
    ApplyAction: URIRef
    Aquarium: URIRef
    ArriveAction: URIRef
    ArtGallery: URIRef
    Article: URIRef
    AskAction: URIRef
    AssessAction: URIRef
    AssignAction: URIRef
    Attorney: URIRef
    Audience: URIRef
    AudioObject: URIRef
    AuthorizeAction: URIRef
    AutoBodyShop: URIRef
    AutoDealer: URIRef
    AutoPartsStore: URIRef
    AutoRental: URIRef
    AutoRepair: URIRef
    AutoWash: URIRef
    AutomatedTeller: URIRef
    AutomotiveBusiness: URIRef
    Bakery: URIRef
    BankAccount: URIRef
    BankOrCreditUnion: URIRef
    BarOrPub: URIRef
    Barcode: URIRef
    Beach: URIRef
    BeautySalon: URIRef
    BedAndBreakfast: URIRef
    BedDetails: URIRef
    BedType: URIRef
    BefriendAction: URIRef
    BikeStore: URIRef
    Blog: URIRef
    BlogPosting: URIRef
    BoardingPolicyType: URIRef
    BodyOfWater: URIRef
    Book: URIRef
    BookFormatType: URIRef
    BookSeries: URIRef
    BookStore: URIRef
    BookmarkAction: URIRef
    Boolean: URIRef
    BorrowAction: URIRef
    BowlingAlley: URIRef
    Brand: URIRef
    BreadcrumbList: URIRef
    Brewery: URIRef
    Bridge: URIRef
    BroadcastChannel: URIRef
    BroadcastEvent: URIRef
    BroadcastFrequencySpecification: URIRef
    BroadcastService: URIRef
    BuddhistTemple: URIRef
    BusReservation: URIRef
    BusStation: URIRef
    BusStop: URIRef
    BusTrip: URIRef
    BusinessAudience: URIRef
    BusinessEntityType: URIRef
    BusinessEvent: URIRef
    BusinessFunction: URIRef
    BuyAction: URIRef
    CableOrSatelliteService: URIRef
    CafeOrCoffeeShop: URIRef
    Campground: URIRef
    CampingPitch: URIRef
    Canal: URIRef
    CancelAction: URIRef
    Car: URIRef
    Casino: URIRef
    CatholicChurch: URIRef
    Cemetery: URIRef
    CheckAction: URIRef
    CheckInAction: URIRef
    CheckOutAction: URIRef
    CheckoutPage: URIRef
    ChildCare: URIRef
    ChildrensEvent: URIRef
    ChooseAction: URIRef
    Church: URIRef
    City: URIRef
    CityHall: URIRef
    CivicStructure: URIRef
    ClaimReview: URIRef
    Clip: URIRef
    ClothingStore: URIRef
    Code: URIRef
    CollectionPage: URIRef
    CollegeOrUniversity: URIRef
    ComedyClub: URIRef
    ComedyEvent: URIRef
    Comment: URIRef
    CommentAction: URIRef
    CommunicateAction: URIRef
    CompoundPriceSpecification: URIRef
    ComputerLanguage: URIRef
    ComputerStore: URIRef
    ConfirmAction: URIRef
    ConsumeAction: URIRef
    ContactPage: URIRef
    ContactPoint: URIRef
    ContactPointOption: URIRef
    Continent: URIRef
    ControlAction: URIRef
    ConvenienceStore: URIRef
    Conversation: URIRef
    CookAction: URIRef
    Corporation: URIRef
    Country: URIRef
    Course: URIRef
    CourseInstance: URIRef
    Courthouse: URIRef
    CreateAction: URIRef
    CreativeWork: URIRef
    CreativeWorkSeason: URIRef
    CreativeWorkSeries: URIRef
    CreditCard: URIRef
    Crematorium: URIRef
    CurrencyConversionService: URIRef
    DanceEvent: URIRef
    DanceGroup: URIRef
    DataCatalog: URIRef
    DataDownload: URIRef
    DataFeed: URIRef
    DataFeedItem: URIRef
    DataType: URIRef
    Dataset: URIRef
    Date: URIRef
    DateTime: URIRef
    DatedMoneySpecification: URIRef
    DayOfWeek: URIRef
    DaySpa: URIRef
    DeactivateAction: URIRef
    DefenceEstablishment: URIRef
    DeleteAction: URIRef
    DeliveryChargeSpecification: URIRef
    DeliveryEvent: URIRef
    DeliveryMethod: URIRef
    Demand: URIRef
    Dentist: URIRef
    DepartAction: URIRef
    DepartmentStore: URIRef
    DepositAccount: URIRef
    DigitalDocument: URIRef
    DigitalDocumentPermission: URIRef
    DigitalDocumentPermissionType: URIRef
    DisagreeAction: URIRef
    DiscoverAction: URIRef
    DiscussionForumPosting: URIRef
    DislikeAction: URIRef
    Distance: URIRef
    Distillery: URIRef
    DonateAction: URIRef
    DownloadAction: URIRef
    DrawAction: URIRef
    DrinkAction: URIRef
    DriveWheelConfigurationValue: URIRef
    DryCleaningOrLaundry: URIRef
    Duration: URIRef
    EatAction: URIRef
    EducationEvent: URIRef
    EducationalAudience: URIRef
    EducationalOrganization: URIRef
    Electrician: URIRef
    ElectronicsStore: URIRef
    ElementarySchool: URIRef
    EmailMessage: URIRef
    Embassy: URIRef
    EmergencyService: URIRef
    EmployeeRole: URIRef
    EmployerAggregateRating: URIRef
    EmploymentAgency: URIRef
    EndorseAction: URIRef
    EndorsementRating: URIRef
    Energy: URIRef
    EngineSpecification: URIRef
    EntertainmentBusiness: URIRef
    EntryPoint: URIRef
    Enumeration: URIRef
    Episode: URIRef
    Event: URIRef
    EventReservation: URIRef
    EventStatusType: URIRef
    EventVenue: URIRef
    ExerciseAction: URIRef
    ExerciseGym: URIRef
    ExhibitionEvent: URIRef
    FAQPage: URIRef
    FMRadioChannel: URIRef
    FastFoodRestaurant: URIRef
    Festival: URIRef
    FilmAction: URIRef
    FinancialProduct: URIRef
    FinancialService: URIRef
    FindAction: URIRef
    FireStation: URIRef
    Flight: URIRef
    FlightReservation: URIRef
    Float: URIRef
    Florist: URIRef
    FollowAction: URIRef
    FoodEstablishment: URIRef
    FoodEstablishmentReservation: URIRef
    FoodEvent: URIRef
    FoodService: URIRef
    FurnitureStore: URIRef
    Game: URIRef
    GamePlayMode: URIRef
    GameServer: URIRef
    GameServerStatus: URIRef
    GardenStore: URIRef
    GasStation: URIRef
    GatedResidenceCommunity: URIRef
    GenderType: URIRef
    GeneralContractor: URIRef
    GeoCircle: URIRef
    GeoCoordinates: URIRef
    GeoShape: URIRef
    GiveAction: URIRef
    GolfCourse: URIRef
    GovernmentBuilding: URIRef
    GovernmentOffice: URIRef
    GovernmentOrganization: URIRef
    GovernmentPermit: URIRef
    GovernmentService: URIRef
    GroceryStore: URIRef
    HVACBusiness: URIRef
    HairSalon: URIRef
    HardwareStore: URIRef
    HealthAndBeautyBusiness: URIRef
    HealthClub: URIRef
    HighSchool: URIRef
    HinduTemple: URIRef
    HobbyShop: URIRef
    HomeAndConstructionBusiness: URIRef
    HomeGoodsStore: URIRef
    Hospital: URIRef
    Hostel: URIRef
    Hotel: URIRef
    HotelRoom: URIRef
    House: URIRef
    HousePainter: URIRef
    HowTo: URIRef
    HowToDirection: URIRef
    HowToItem: URIRef
    HowToSection: URIRef
    HowToStep: URIRef
    HowToSupply: URIRef
    HowToTip: URIRef
    HowToTool: URIRef
    IceCreamShop: URIRef
    IgnoreAction: URIRef
    ImageGallery: URIRef
    ImageObject: URIRef
    IndividualProduct: URIRef
    InformAction: URIRef
    InsertAction: URIRef
    InstallAction: URIRef
    InsuranceAgency: URIRef
    Intangible: URIRef
    Integer: URIRef
    InteractAction: URIRef
    InteractionCounter: URIRef
    InternetCafe: URIRef
    InvestmentOrDeposit: URIRef
    InviteAction: URIRef
    Invoice: URIRef
    ItemAvailability: URIRef
    ItemList: URIRef
    ItemListOrderType: URIRef
    ItemPage: URIRef
    JewelryStore: URIRef
    JobPosting: URIRef
    JoinAction: URIRef
    LakeBodyOfWater: URIRef
    Landform: URIRef
    LandmarksOrHistoricalBuildings: URIRef
    Language: URIRef
    LeaveAction: URIRef
    LegalService: URIRef
    LegislativeBuilding: URIRef
    LendAction: URIRef
    Library: URIRef
    LikeAction: URIRef
    LiquorStore: URIRef
    ListItem: URIRef
    ListenAction: URIRef
    LiteraryEvent: URIRef
    LiveBlogPosting: URIRef
    LoanOrCredit: URIRef
    LocalBusiness: URIRef
    LocationFeatureSpecification: URIRef
    LockerDelivery: URIRef
    Locksmith: URIRef
    LodgingBusiness: URIRef
    LodgingReservation: URIRef
    LoseAction: URIRef
    Map: URIRef
    MapCategoryType: URIRef
    MarryAction: URIRef
    Mass: URIRef
    MediaGallery: URIRef
    MediaObject: URIRef
    MediaSubscription: URIRef
    MedicalOrganization: URIRef
    MeetingRoom: URIRef
    MensClothingStore: URIRef
    Menu: URIRef
    MenuItem: URIRef
    MenuSection: URIRef
    Message: URIRef
    MiddleSchool: URIRef
    MobileApplication: URIRef
    MobilePhoneStore: URIRef
    MonetaryAmount: URIRef
    MonetaryAmountDistribution: URIRef
    Mosque: URIRef
    Motel: URIRef
    MotorcycleDealer: URIRef
    MotorcycleRepair: URIRef
    Mountain: URIRef
    MoveAction: URIRef
    Movie: URIRef
    MovieClip: URIRef
    MovieRentalStore: URIRef
    MovieSeries: URIRef
    MovieTheater: URIRef
    MovingCompany: URIRef
    Museum: URIRef
    MusicAlbum: URIRef
    MusicAlbumProductionType: URIRef
    MusicAlbumReleaseType: URIRef
    MusicComposition: URIRef
    MusicEvent: URIRef
    MusicGroup: URIRef
    MusicPlaylist: URIRef
    MusicRecording: URIRef
    MusicRelease: URIRef
    MusicReleaseFormatType: URIRef
    MusicStore: URIRef
    MusicVenue: URIRef
    MusicVideoObject: URIRef
    NGO: URIRef
    NailSalon: URIRef
    NewsArticle: URIRef
    NightClub: URIRef
    Notary: URIRef
    NoteDigitalDocument: URIRef
    Number: URIRef
    NutritionInformation: URIRef
    Occupation: URIRef
    OceanBodyOfWater: URIRef
    Offer: URIRef
    OfferCatalog: URIRef
    OfferItemCondition: URIRef
    OfficeEquipmentStore: URIRef
    OnDemandEvent: URIRef
    OpeningHoursSpecification: URIRef
    Order: URIRef
    OrderAction: URIRef
    OrderItem: URIRef
    OrderStatus: URIRef
    Organization: URIRef
    OrganizationRole: URIRef
    OrganizeAction: URIRef
    OutletStore: URIRef
    OwnershipInfo: URIRef
    PaintAction: URIRef
    Painting: URIRef
    ParcelDelivery: URIRef
    ParcelService: URIRef
    ParentAudience: URIRef
    Park: URIRef
    ParkingFacility: URIRef
    PawnShop: URIRef
    PayAction: URIRef
    PaymentCard: URIRef
    PaymentChargeSpecification: URIRef
    PaymentMethod: URIRef
    PaymentService: URIRef
    PaymentStatusType: URIRef
    PeopleAudience: URIRef
    PerformAction: URIRef
    PerformanceRole: URIRef
    PerformingArtsTheater: URIRef
    PerformingGroup: URIRef
    Periodical: URIRef
    Permit: URIRef
    Person: URIRef
    PetStore: URIRef
    Pharmacy: URIRef
    Photograph: URIRef
    PhotographAction: URIRef
    Physician: URIRef
    Place: URIRef
    PlaceOfWorship: URIRef
    PlanAction: URIRef
    PlayAction: URIRef
    Playground: URIRef
    Plumber: URIRef
    PoliceStation: URIRef
    Pond: URIRef
    PostOffice: URIRef
    PostalAddress: URIRef
    PreOrderAction: URIRef
    PrependAction: URIRef
    Preschool: URIRef
    PresentationDigitalDocument: URIRef
    PriceSpecification: URIRef
    Product: URIRef
    ProductModel: URIRef
    ProfessionalService: URIRef
    ProfilePage: URIRef
    ProgramMembership: URIRef
    PropertyValue: URIRef
    PropertyValueSpecification: URIRef
    PublicSwimmingPool: URIRef
    PublicationEvent: URIRef
    PublicationIssue: URIRef
    PublicationVolume: URIRef
    QAPage: URIRef
    QualitativeValue: URIRef
    QuantitativeValue: URIRef
    QuantitativeValueDistribution: URIRef
    Quantity: URIRef
    Question: URIRef
    QuoteAction: URIRef
    RVPark: URIRef
    RadioChannel: URIRef
    RadioClip: URIRef
    RadioEpisode: URIRef
    RadioSeason: URIRef
    RadioSeries: URIRef
    RadioStation: URIRef
    Rating: URIRef
    ReactAction: URIRef
    ReadAction: URIRef
    RealEstateAgent: URIRef
    ReceiveAction: URIRef
    Recipe: URIRef
    RecyclingCenter: URIRef
    RegisterAction: URIRef
    RejectAction: URIRef
    RentAction: URIRef
    RentalCarReservation: URIRef
    ReplaceAction: URIRef
    ReplyAction: URIRef
    Report: URIRef
    Reservation: URIRef
    ReservationPackage: URIRef
    ReservationStatusType: URIRef
    ReserveAction: URIRef
    Reservoir: URIRef
    Residence: URIRef
    Resort: URIRef
    Restaurant: URIRef
    RestrictedDiet: URIRef
    ResumeAction: URIRef
    ReturnAction: URIRef
    Review: URIRef
    ReviewAction: URIRef
    RiverBodyOfWater: URIRef
    Role: URIRef
    RoofingContractor: URIRef
    Room: URIRef
    RsvpAction: URIRef
    RsvpResponseType: URIRef
    SaleEvent: URIRef
    ScheduleAction: URIRef
    ScholarlyArticle: URIRef
    School: URIRef
    ScreeningEvent: URIRef
    Sculpture: URIRef
    SeaBodyOfWater: URIRef
    SearchAction: URIRef
    SearchResultsPage: URIRef
    Season: URIRef
    Seat: URIRef
    SelfStorage: URIRef
    SellAction: URIRef
    SendAction: URIRef
    Series: URIRef
    Service: URIRef
    ServiceChannel: URIRef
    ShareAction: URIRef
    ShoeStore: URIRef
    ShoppingCenter: URIRef
    SingleFamilyResidence: URIRef
    SiteNavigationElement: URIRef
    SkiResort: URIRef
    SocialEvent: URIRef
    SocialMediaPosting: URIRef
    SoftwareApplication: URIRef
    SoftwareSourceCode: URIRef
    SomeProducts: URIRef
    SpeakableSpecification: URIRef
    Specialty: URIRef
    SportingGoodsStore: URIRef
    SportsActivityLocation: URIRef
    SportsClub: URIRef
    SportsEvent: URIRef
    SportsOrganization: URIRef
    SportsTeam: URIRef
    SpreadsheetDigitalDocument: URIRef
    StadiumOrArena: URIRef
    State: URIRef
    SteeringPositionValue: URIRef
    Store: URIRef
    StructuredValue: URIRef
    SubscribeAction: URIRef
    SubwayStation: URIRef
    Suite: URIRef
    SuspendAction: URIRef
    Synagogue: URIRef
    TVClip: URIRef
    TVEpisode: URIRef
    TVSeason: URIRef
    TVSeries: URIRef
    Table: URIRef
    TakeAction: URIRef
    TattooParlor: URIRef
    Taxi: URIRef
    TaxiReservation: URIRef
    TaxiService: URIRef
    TaxiStand: URIRef
    TechArticle: URIRef
    TelevisionChannel: URIRef
    TelevisionStation: URIRef
    TennisComplex: URIRef
    Text: URIRef
    TextDigitalDocument: URIRef
    TheaterEvent: URIRef
    TheaterGroup: URIRef
    Thing: URIRef
    Ticket: URIRef
    TieAction: URIRef
    Time: URIRef
    TipAction: URIRef
    TireShop: URIRef
    TouristAttraction: URIRef
    TouristInformationCenter: URIRef
    ToyStore: URIRef
    TrackAction: URIRef
    TradeAction: URIRef
    TrainReservation: URIRef
    TrainStation: URIRef
    TrainTrip: URIRef
    TransferAction: URIRef
    TravelAction: URIRef
    TravelAgency: URIRef
    Trip: URIRef
    TypeAndQuantityNode: URIRef
    URL: URIRef
    UnRegisterAction: URIRef
    UnitPriceSpecification: URIRef
    UpdateAction: URIRef
    UseAction: URIRef
    UserBlocks: URIRef
    UserCheckins: URIRef
    UserComments: URIRef
    UserDownloads: URIRef
    UserInteraction: URIRef
    UserLikes: URIRef
    UserPageVisits: URIRef
    UserPlays: URIRef
    UserPlusOnes: URIRef
    UserTweets: URIRef
    Vehicle: URIRef
    VideoGallery: URIRef
    VideoGame: URIRef
    VideoGameClip: URIRef
    VideoGameSeries: URIRef
    VideoObject: URIRef
    ViewAction: URIRef
    VisualArtsEvent: URIRef
    VisualArtwork: URIRef
    Volcano: URIRef
    VoteAction: URIRef
    WPAdBlock: URIRef
    WPFooter: URIRef
    WPHeader: URIRef
    WPSideBar: URIRef
    WantAction: URIRef
    WarrantyPromise: URIRef
    WarrantyScope: URIRef
    WatchAction: URIRef
    Waterfall: URIRef
    WearAction: URIRef
    WebApplication: URIRef
    WebPage: URIRef
    WebPageElement: URIRef
    WebSite: URIRef
    WholesaleStore: URIRef
    WinAction: URIRef
    Winery: URIRef
    WorkersUnion: URIRef
    WriteAction: URIRef
    Zoo: URIRef



import inspect
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, cast

import schema_salad.metaschema as cg_metaschema
from schema_salad import codegen
from schema_salad.avro.schema import Names
from schema_salad.schema import load_schema

from .test_java_codegen import cwl_file_uri, metaschema_file_uri


def test_cwl_gen(tmp_path: Path) -> None:
    src_target = tmp_path / "src.py"
    python_codegen(cwl_file_uri, src_target)
    assert os.path.exists(src_target)
    with open(src_target) as f:
        assert "class Workflow(Process)" in f.read()


def test_meta_schema_gen(tmp_path: Path) -> None:
    src_target = tmp_path / "src.py"
    python_codegen(metaschema_file_uri, src_target)
    assert os.path.exists(src_target)
    with open(src_target) as f:
        assert "class RecordSchema(Savable):" in f.read()


def test_meta_schema_gen_up_to_date(tmp_path: Path) -> None:
    src_target = tmp_path / "src.py"
    python_codegen(metaschema_file_uri, src_target)
    assert os.path.exists(src_target)
    with open(src_target) as f:
        assert f.read() == inspect.getsource(cg_metaschema)


def python_codegen(
    file_uri: str,
    target: Path,
    parser_info: Optional[str] = None,
    package: Optional[str] = None,
) -> None:
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(
        file_uri
    )
    assert isinstance(avsc_names, Names)
    schema_raw_doc = metaschema_loader.fetch(file_uri)
    schema_doc, schema_metadata = metaschema_loader.resolve_all(
        schema_raw_doc, file_uri
    )
    codegen.codegen(
        "python",
        cast(List[Dict[str, Any]], schema_doc),
        schema_metadata,
        document_loader,
        target=str(target),
        parser_info=parser_info,
        package=package,
    )


def test_default_parser_info(tmp_path: Path) -> None:
    src_target = tmp_path / "src.py"
    python_codegen(metaschema_file_uri, src_target)
    assert os.path.exists(src_target)
    with open(src_target) as f:
        assert 'def parser_info() -> str:\n    return "org.w3id.cwl.salad"' in f.read()


def test_parser_info(tmp_path: Path) -> None:
    src_target = tmp_path / "src.py"
    python_codegen(metaschema_file_uri, src_target, parser_info="cwl")
    assert os.path.exists(src_target)
    with open(src_target) as f:
        assert 'def parser_info() -> str:\n    return "cwl"' in f.read()


def test_use_of_package_for_parser_info(tmp_path: Path) -> None:
    src_target = tmp_path / "src.py"
    python_codegen(metaschema_file_uri, src_target, package="cwl")
    assert os.path.exists(src_target)
    with open(src_target) as f:
        assert 'def parser_info() -> str:\n    return "cwl"' in f.read()



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class SKOS(DefinedNamespace):
    altLabel: URIRef
    broadMatch: URIRef
    broader: URIRef
    broaderTransitive: URIRef
    changeNote: URIRef
    closeMatch: URIRef
    definition: URIRef
    editorialNote: URIRef
    exactMatch: URIRef
    example: URIRef
    hasTopConcept: URIRef
    hiddenLabel: URIRef
    historyNote: URIRef
    inScheme: URIRef
    mappingRelation: URIRef
    member: URIRef
    memberList: URIRef
    narrowMatch: URIRef
    narrower: URIRef
    narrowerTransitive: URIRef
    notation: URIRef
    note: URIRef
    prefLabel: URIRef
    related: URIRef
    relatedMatch: URIRef
    scopeNote: URIRef
    semanticRelation: URIRef
    topConceptOf: URIRef
    Collection: URIRef
    Concept: URIRef
    ConceptScheme: URIRef
    OrderedCollection: URIRef



"""Avro related tests."""
from schema_salad.avro.schema import Names
from schema_salad.schema import load_schema

from .util import get_data


def test_avro_loading() -> None:
    """Confirm conversion of SALAD style names to avro."""
    path = get_data("tests/test_schema/avro_naming.yml")
    assert path
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(path)
    assert isinstance(avsc_names, Names)
    assert avsc_names.get_name("com.example.derived_schema.ExtendedThing", None)



from typing import Any, Callable, Union

from rdflib.term import Node as Node
from rdflib.term import URIRef as URIRef

ZeroOrMore: str
OneOrMore: str
ZeroOrOne: str

class Path:
    __or__: Callable[[Path, Union[URIRef, Path]], AlternativePath]
    __invert__: Callable[[Path], InvPath]
    __neg__: Callable[[Path], NegatedPath]
    __truediv__: Callable[[Path, Union[URIRef, Path]], SequencePath]
    __mul__: Callable[[Path, str], MulPath]
    def eval(self, graph, subj: Any | None = ..., obj: Any | None = ...) -> None: ...
    def __hash__(self): ...
    def __eq__(self, other): ...
    def __lt__(self, other): ...
    def __le__(self, other): ...
    def __ne__(self, other): ...
    def __gt__(self, other): ...
    def __ge__(self, other): ...

class InvPath(Path):
    arg: Any
    def __init__(self, arg) -> None: ...
    def eval(self, graph, subj: Any | None = ..., obj: Any | None = ...) -> None: ...
    def n3(self): ...

class SequencePath(Path):
    args: Any
    def __init__(self, *args) -> None: ...
    def eval(self, graph, subj: Any | None = ..., obj: Any | None = ...): ...
    def n3(self): ...

class AlternativePath(Path):
    args: Any
    def __init__(self, *args) -> None: ...
    def eval(self, graph, subj: Any | None = ..., obj: Any | None = ...) -> None: ...
    def n3(self): ...

class MulPath(Path):
    path: Any
    mod: Any
    zero: bool
    more: bool
    def __init__(self, path, mod) -> None: ...
    def eval(
        self, graph, subj: Any | None = ..., obj: Any | None = ..., first: bool = ...
    ) -> None: ...
    def n3(self): ...

class NegatedPath(Path):
    args: Any
    def __init__(self, arg) -> None: ...
    def eval(self, graph, subj: Any | None = ..., obj: Any | None = ...) -> None: ...
    def n3(self): ...

class PathList(list): ...

def path_alternative(self, other): ...
def path_sequence(self, other): ...
def evalPath(graph, t): ...
def mul_path(p, mul): ...
def inv_path(p): ...
def neg_path(p): ...



from typing import Any

class Serializer:
    store: Any
    encoding: str
    base: Any
    def __init__(self, store) -> None: ...
    def serialize(
        self, stream, base: Any | None = ..., encoding: Any | None = ..., **args
    ) -> None: ...
    def relativize(self, uri): ...



# Stubs for cachecontrol.caches (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

from .file_cache import FileCache as FileCache

# from .redis_cache import RedisCache as RedisCache

notice = ...  # type: Any



# Stubs for cachecontrol.filewrapper (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

class CallbackFileWrapper:
    def __init__(self, fp, callback): ...
    def __getattr__(self, name): ...
    def read(self, amt=None): ...



from typing import Any, Iterator, List, Optional, Tuple, Union, overload

from rdflib.term import BNode, Identifier, Node

class Graph(Node):
    base: Any = ...
    context_aware: bool = ...
    formula_aware: bool = ...
    default_union: bool = ...
    def __init__(
        self,
        store: str = ...,
        identifier: Optional[Any] = ...,
        namespace_manager: Optional[Any] = ...,
        base: Optional[Any] = ...,
    ) -> None: ...
    store: Any = ...
    identifier: Any = ...
    namespace_manager: Any = ...
    def toPython(self): ...
    def destroy(self, configuration: Any) -> None: ...
    def commit(self) -> None: ...
    def rollback(self) -> None: ...
    def open(self, configuration: Any, create: bool = ...): ...
    def close(self, commit_pending_transaction: bool = ...) -> None: ...
    def add(self, triple: Any) -> None: ...
    def addN(self, quads: Any) -> None: ...
    def remove(self, triple: Any) -> None: ...
    def triples(
        self,
        triple: Tuple[
            Optional[Union[str, Identifier]],
            Optional[Union[str, Identifier]],
            Optional[Identifier],
        ],
    ) -> Iterator[Tuple[Identifier, Identifier, Identifier]]: ...
    def __getitem__(self, item: Any): ...
    def __len__(self): ...
    def __iter__(self) -> Any: ...
    def __contains__(self, triple: Any): ...
    def __hash__(self) -> Any: ...
    def __cmp__(self, other: Any): ...
    def __eq__(self, other: Any) -> Any: ...
    def __lt__(self, other: Any) -> Any: ...
    def __le__(self, other: Any) -> Any: ...
    def __gt__(self, other: Any) -> Any: ...
    def __ge__(self, other: Any) -> Any: ...
    def __iadd__(self, other: Any): ...
    def __isub__(self, other: Any): ...
    def __add__(self, other: Any): ...
    def __mul__(self, other: Any): ...
    def __sub__(self, other: Any): ...
    def __xor__(self, other: Any): ...
    __or__: Any = ...
    __and__: Any = ...
    def set(self, triple: Any) -> None: ...
    def subjects(
        self, predicate: Optional[Any] = ..., object: Optional[Any] = ...
    ) -> List[str]: ...
    def predicates(
        self, subject: Optional[Any] = ..., object: Optional[Any] = ...
    ) -> None: ...
    def objects(
        self, subject: Optional[Any] = ..., predicate: Optional[Any] = ...
    ) -> List[str]: ...
    def subject_predicates(self, object: Optional[Any] = ...) -> None: ...
    def subject_objects(self, predicate: Optional[Any] = ...) -> None: ...
    def predicate_objects(self, subject: Optional[Any] = ...) -> None: ...
    def triples_choices(self, triple: Any, context: Optional[Any] = ...) -> None: ...
    def value(
        self,
        subject: Optional[Any] = ...,
        predicate: Any = ...,
        object: Optional[Any] = ...,
        default: Optional[Any] = ...,
        any: bool = ...,
    ): ...
    def label(self, subject: Any, default: str = ...): ...
    def preferredLabel(
        self,
        subject: Any,
        lang: Optional[Any] = ...,
        default: Optional[Any] = ...,
        labelProperties: Any = ...,
    ): ...
    def comment(self, subject: Any, default: str = ...): ...
    def items(self, list: Any) -> None: ...
    def transitiveClosure(
        self, func: Any, arg: Any, seen: Optional[Any] = ...
    ) -> None: ...
    def transitive_objects(
        self, subject: Any, property: Any, remember: Optional[Any] = ...
    ) -> None: ...
    def transitive_subjects(
        self, predicate: Any, object: Any, remember: Optional[Any] = ...
    ) -> None: ...
    def seq(self, subject: Any): ...
    def qname(self, uri: Any): ...
    def compute_qname(self, uri: Any, generate: bool = ...): ...
    def bind(
        self, prefix: Any, namespace: Any, override: bool = ..., replace: bool = ...
    ): ...
    def namespaces(self) -> None: ...
    def absolutize(self, uri: Any, defrag: int = ...): ...
    def serialize(
        self,
        destination: Optional[Any] = ...,
        format: str = ...,
        base: Optional[Any] = ...,
        encoding: Optional[Any] = ...,
        **args: Any
    ): ...
    def parse(
        self,
        source: Optional[Any] = ...,
        publicID: Optional[Any] = ...,
        format: Optional[str] = ...,
        location: Optional[Any] = ...,
        file: Optional[Any] = ...,
        data: Optional[Any] = ...,
        **args: Any
    ): ...
    def load(
        self, source: Any, publicID: Optional[Any] = ..., format: str = ...
    ) -> None: ...
    def query(
        self,
        query_object: Any,
        processor: str = ...,
        result: str = ...,
        initNs: Optional[Any] = ...,
        initBindings: Optional[Any] = ...,
        use_store_provided: bool = ...,
        **kwargs: Any
    ): ...
    def update(
        self,
        update_object: Any,
        processor: str = ...,
        initNs: Optional[Any] = ...,
        initBindings: Optional[Any] = ...,
        use_store_provided: bool = ...,
        **kwargs: Any
    ): ...
    def n3(self): ...
    def __reduce__(self): ...
    def isomorphic(self, other: Any): ...
    def connected(self): ...
    def all_nodes(self): ...
    def collection(self, identifier: Any): ...
    def resource(self, identifier: Any): ...
    def skolemize(
        self,
        new_graph: Optional[Any] = ...,
        bnode: Optional[Any] = ...,
        authority: Optional[Any] = ...,
        basepath: Optional[Any] = ...,
    ): ...
    def de_skolemize(
        self, new_graph: Optional[Any] = ..., uriref: Optional[Any] = ...
    ): ...

class ConjunctiveGraph(Graph):
    context_aware: bool = ...
    default_union: bool = ...
    default_context: Any = ...
    def __init__(
        self,
        store: str = ...,
        identifier: Optional[Any] = ...,
        default_graph_base: Optional[Any] = ...,
    ) -> None: ...
    def __contains__(self, triple_or_quad: Any): ...
    def add(self, triple_or_quad: Any) -> None: ...
    def addN(self, quads: Any) -> None: ...
    def remove(self, triple_or_quad: Any) -> None: ...
    # def triples(self, triple_or_quad: Tuple[Optional[Union[str, BNode]], Optional[Union[str, BNode]], Optional[BNode]], context: Tuple[Optional[Union[str, BNode]], Optional[Union[str, BNode]], Optional[BNode]]) -> Iterator[Tuple[Identifier, Identifier, Identifier]]: ...
    def quads(self, triple_or_quad: Optional[Any] = ...) -> None: ...
    def triples_choices(self, triple: Any, context: Optional[Any] = ...) -> None: ...
    def __len__(self): ...
    def contexts(self, triple: Optional[Any] = ...) -> None: ...
    def get_context(
        self, identifier: Any, quoted: bool = ..., base: Optional[Any] = ...
    ): ...
    def remove_context(self, context: Any) -> None: ...
    def context_id(self, uri: Any, context_id: Optional[Any] = ...): ...
    def parse(
        self,
        source: Optional[Any] = ...,
        publicID: Optional[Any] = ...,
        format: Optional[str] = ...,
        location: Optional[Any] = ...,
        file: Optional[Any] = ...,
        data: Optional[Any] = ...,
        **args: Any
    ): ...
    def __reduce__(self): ...

class Dataset(ConjunctiveGraph):
    __doc__: str = ...
    default_context: Any = ...
    default_union: Any = ...
    def __init__(
        self,
        store: str = ...,
        default_union: bool = ...,
        default_graph_base: Optional[Any] = ...,
    ) -> None: ...
    def graph(self, identifier: Optional[Any] = ..., base: Optional[Any] = ...): ...
    def parse(
        self,
        source: Optional[Any] = ...,
        publicID: Optional[Any] = ...,
        format: Optional[str] = ...,
        location: Optional[Any] = ...,
        file: Optional[Any] = ...,
        data: Optional[Any] = ...,
        **args: Any
    ): ...
    def add_graph(self, g: Any): ...
    def remove_graph(self, g: Any) -> None: ...
    def contexts(self, triple: Optional[Any] = ...) -> None: ...
    graphs: Any = ...

class QuotedGraph(Graph):
    def __init__(self, store: Any, identifier: Any) -> None: ...
    def add(self, triple: Any) -> None: ...
    def addN(self, quads: Any) -> None: ...
    def n3(self): ...
    def __reduce__(self): ...

class Seq:
    def __init__(self, graph: Any, subject: Any) -> None: ...
    def toPython(self): ...
    def __iter__(self) -> Any: ...
    def __len__(self): ...
    def __getitem__(self, index: Any): ...

class ModificationException(Exception):
    def __init__(self) -> None: ...

class UnSupportedAggregateOperation(Exception):
    def __init__(self) -> None: ...

class ReadOnlyGraphAggregate(ConjunctiveGraph):
    graphs: Any = ...
    def __init__(self, graphs: Any, store: str = ...) -> None: ...
    def destroy(self, configuration: Any) -> None: ...
    def commit(self) -> None: ...
    def rollback(self) -> None: ...
    def open(self, configuration: Any, create: bool = ...) -> None: ...
    def add(self, triple: Any) -> None: ...
    def addN(self, quads: Any) -> None: ...
    def remove(self, triple: Any) -> None: ...
    def triples(
        self,
        triple: Tuple[
            Optional[Union[str, Identifier]],
            Optional[Union[str, Identifier]],
            Optional[Identifier],
        ],
    ) -> Iterator[Tuple[Identifier, Identifier, Identifier]]: ...
    def __contains__(self, triple_or_quad: Any): ...
    def __len__(self): ...
    def __hash__(self) -> Any: ...
    def __cmp__(self, other: Any): ...
    def __iadd__(self, other: Any) -> None: ...
    def __isub__(self, other: Any) -> None: ...
    def triples_choices(self, triple: Any, context: Optional[Any] = ...) -> None: ...
    def qname(self, uri: Any): ...
    def compute_qname(self, uri: Any, generate: bool = ...): ...
    def namespaces(self) -> None: ...
    def absolutize(self, uri: Any, defrag: int = ...) -> None: ...
    def n3(self) -> None: ...



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class SOSA(DefinedNamespace):
    ActuatableProperty: URIRef
    Actuation: URIRef
    Actuator: URIRef
    FeatureOfInterest: URIRef
    ObservableProperty: URIRef
    Observation: URIRef
    Platform: URIRef
    Procedure: URIRef
    Result: URIRef
    Sample: URIRef
    Sampler: URIRef
    Sampling: URIRef
    Sensor: URIRef
    hasSimpleResult: URIRef
    resultTime: URIRef
    actsOnProperty: URIRef
    hasFeatureOfInterest: URIRef
    hasResult: URIRef
    hasSample: URIRef
    hosts: URIRef
    isActedOnBy: URIRef
    isFeatureOfInterestOf: URIRef
    isHostedBy: URIRef
    isObservedBy: URIRef
    isResultOf: URIRef
    isSampleOf: URIRef
    madeActuation: URIRef
    madeByActuator: URIRef
    madeBySampler: URIRef
    madeBySensor: URIRef
    madeObservation: URIRef
    madeSampling: URIRef
    observedProperty: URIRef
    observes: URIRef
    phenomenonTime: URIRef
    usedProcedure: URIRef



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class RDFS(DefinedNamespace):
    comment: URIRef
    domain: URIRef
    isDefinedBy: URIRef
    label: URIRef
    member: URIRef
    range: URIRef
    seeAlso: URIRef
    subClassOf: URIRef
    subPropertyOf: URIRef
    Class: URIRef
    Container: URIRef
    ContainerMembershipProperty: URIRef
    Datatype: URIRef
    Literal: URIRef
    Resource: URIRef



"""
Ensure codegen-produced parsers accept $schemas directives

run individually as py.test -k test_cwl11
"""

import os
import shutil
import tarfile
from typing import Any, Dict, Generator, Tuple, Union

import pytest
import requests
from _pytest.tmpdir import TempPathFactory

from schema_salad.avro.schema import Names, SchemaParseException
from schema_salad.ref_resolver import Loader
from schema_salad.schema import load_and_validate, load_schema

from .util import get_data

test_dir_name = "tests/"

SchemaType = Tuple[Loader, Union[Names, SchemaParseException], Dict[str, Any], Loader]


@pytest.fixture(scope="session")
def cwl_v1_2_schema(
    tmp_path_factory: TempPathFactory,
) -> Generator[SchemaType, None, None]:
    tmp_path = tmp_path_factory.mktemp("cwl_v1_2_schema")
    with requests.get(
        "https://github.com/common-workflow-language/cwl-v1.2/archive/v1.2.0.tar.gz",
        stream=True,
    ).raw as specfileobj:
        tf = tarfile.open(fileobj=specfileobj)
        tf.extractall(path=tmp_path)  # this becomes cwl-v1.2-1.2.0
    path = str(tmp_path / "cwl-v1.2-1.2.0/CommonWorkflowLanguage.yml")
    yield load_schema(path)
    shutil.rmtree(os.path.join(tmp_path))


def load_cwl(cwl_v1_2_schema: SchemaType, src: str) -> Tuple[Any, Dict[str, Any]]:
    (document_loader, avsc_names, schema_metadata, metaschema_loader) = cwl_v1_2_schema
    path = get_data(test_dir_name + src)
    assert path
    assert isinstance(avsc_names, Names)
    res = load_and_validate(document_loader, avsc_names, path, True)
    return res


def test_secondaryFiles(cwl_v1_2_schema: SchemaType) -> None:
    """secondaryFiles"""
    res = load_cwl(
        cwl_v1_2_schema,
        src="test_real_cwl/bio-cwl-tools/picard_CreateSequenceDictionary.cwl",
    )
    print(f"the res:{res}")


def test_outputBinding(cwl_v1_2_schema: SchemaType) -> None:
    """secondaryFiles"""
    res = load_cwl(
        cwl_v1_2_schema, src="test_real_cwl/bio-cwl-tools/bamtools_stats.cwl"
    )
    print(f"the res:{res}")



from typing import Any

class Container:
    graph: Any
    uri: Any
    def __init__(self, graph, uri, seq=..., rtype: str = ...) -> None: ...
    def n3(self): ...
    def __len__(self): ...
    def type_of_conatiner(self): ...
    def index(self, item): ...
    def __getitem__(self, key): ...
    def __setitem__(self, key, value) -> None: ...
    def __delitem__(self, key) -> None: ...
    def items(self): ...
    def end(self): ...
    def append(self, item): ...
    def append_multiple(self, other): ...
    def clear(self): ...

class Bag(Container):
    def __init__(self, graph, uri, seq=...) -> None: ...

class Alt(Container):
    def __init__(self, graph, uri, seq=...) -> None: ...
    def anyone(self): ...

class Seq(Container):
    def __init__(self, graph, uri, seq=...) -> None: ...
    def add_at_position(self, pos, item): ...

class NoElementException(Exception):
    message: Any
    def __init__(self, message: str = ...) -> None: ...



"""TypeScript code generator for a given schema salad definition."""
import os
import shutil
import string
from io import StringIO
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    MutableMapping,
    MutableSequence,
    Optional,
    Set,
    Union,
)

import pkg_resources

from . import schema, _logger
from .codegen_base import CodeGenBase, TypeDef
from .exceptions import SchemaException
from .schema import shortname
from .java_codegen import _ensure_directory_and_write, _safe_makedirs


def doc_to_doc_string(doc: Optional[str], indent_level: int = 0) -> str:
    """Generate a documentation string from a schema salad doc field."""
    lead = " " + "  " * indent_level + "* "
    if doc:
        doc_str = "\n".join([f"{lead}{line}" for line in doc.split("\n")])
    else:
        doc_str = ""
    return doc_str


_string_type_def = TypeDef(
    name="strtype",
    init="new _PrimitiveLoader(TypeGuards.String)",
    instance_type="string",
)

_int_type_def = TypeDef(
    name="inttype", init="new _PrimitiveLoader(TypeGuards.Int)", instance_type="number"
)

_float_type_def = TypeDef(
    name="floattype",
    init="new _PrimitiveLoader(TypeGuards.Float)",
    instance_type="number",
)
_bool_type_def = TypeDef(
    name="booltype",
    init="new _PrimitiveLoader(TypeGuards.Bool)",
    instance_type="boolean",
)

_null_type_def = TypeDef(
    name="undefinedtype",
    init="new _PrimitiveLoader(TypeGuards.Undefined)",
    instance_type="undefined",
)

_any_type_def = TypeDef(name="anyType", init="new _AnyLoader()", instance_type="any")

prims = {
    "http://www.w3.org/2001/XMLSchema#string": _string_type_def,
    "http://www.w3.org/2001/XMLSchema#int": _int_type_def,
    "http://www.w3.org/2001/XMLSchema#long": _int_type_def,
    "http://www.w3.org/2001/XMLSchema#float": _float_type_def,
    "http://www.w3.org/2001/XMLSchema#double": _float_type_def,
    "http://www.w3.org/2001/XMLSchema#boolean": _bool_type_def,
    "https://w3id.org/cwl/salad#null": _null_type_def,
    "https://w3id.org/cwl/salad#Any": _any_type_def,
    "string": _string_type_def,
    "int": _int_type_def,
    "long": _int_type_def,
    "float": _float_type_def,
    "double": _float_type_def,
    "boolean": _bool_type_def,
    "null": _null_type_def,
    "Any": _any_type_def,
}


class TypeScriptCodeGen(CodeGenBase):
    """Generation of TypeScript code for a given Schema Salad definition."""

    def __init__(
        self, base: str, examples: Optional[str], target: Optional[str], package: str
    ) -> None:
        """Initialize the TypeScript codegen."""
        super().__init__()
        self.target_dir = Path(target or ".").resolve()
        self.main_src_dir = self.target_dir / "src"
        self.test_resources_dir = self.target_dir / "src" / "test" / "data"
        self.package = package
        self.base_uri = base
        self.record_types: Set[str] = set()
        self.modules: Set[str] = set()
        self.id_field = ""
        self.examples = examples

    def prologue(self) -> None:
        """Trigger to generate the prolouge code."""
        for src_dir in [self.main_src_dir]:
            _safe_makedirs(src_dir)

        for primitive in prims.values():
            self.declare_type(primitive)

    @staticmethod
    def safe_name(name: str) -> str:
        """Generate a safe version of the given name."""
        avn = schema.avro_field_name(name)
        if avn.startswith("anon."):
            avn = avn[5:]
        if avn in (
            "class",
            "in",
            "extends",
            "abstract",
            "default",
            "package",
            "arguments",
        ):
            # reserved words
            avn = avn + "_"

        return avn

    def begin_class(
        self,  # pylint: disable=too-many-arguments
        classname: str,
        extends: MutableSequence[str],
        doc: str,
        abstract: bool,
        field_names: MutableSequence[str],
        idfield: str,
        optional_fields: Set[str],
    ) -> None:
        """Produce the header for the given class."""
        self.current_interface = self.safe_name(classname) + "Properties"
        cls = self.safe_name(classname)
        self.current_class = cls
        self.current_class_is_abstract = abstract
        interface_module_name = self.current_interface
        self.current_interface_target_file = (
            self.main_src_dir / f"{interface_module_name}.ts"
        )
        class_module_name = self.current_class
        self.current_class_target_file = self.main_src_dir / f"{class_module_name}.ts"
        self.current_constructor_signature = StringIO()
        self.current_constructor_body = StringIO()
        self.current_loader = StringIO()
        self.current_serializer = StringIO()
        self.current_fieldtypes: Dict[str, TypeDef] = {}
        self.idfield = idfield

        doc_string = f"""
/**
 * Auto-generated interface for {classname}
"""
        if doc:
            doc_string += " *\n"
            doc_string += doc_to_doc_string(doc)
            doc_string += "\n"
        doc_string += " */"

        self.record_types.add(f"{self.current_interface}")
        self.modules.add(interface_module_name)
        with open(self.current_interface_target_file, "w") as f:
            _logger.info("Writing file: %s", self.current_interface_target_file)
            if extends:
                ext = "extends Internal." + ", Internal.".join(
                    self.safe_name(e) + "Properties" for e in extends
                )
            else:
                ext = ""
            f.write(
                """
import * as Internal from './util/Internal'

{docstring}
export interface {cls} {ext} {{
                    """.format(
                    docstring=doc_string,
                    cls=f"{self.current_interface}",
                    ext=ext,
                )
            )
        if self.current_class_is_abstract:
            return

        self.record_types.add(cls)
        self.modules.add(class_module_name)
        with open(self.current_interface_target_file, "a") as f:
            f.write(
                """
  extensionFields?: Internal.Dictionary<any>
"""
            )
        doc_string = f"""
/**
 * Auto-generated class implementation for {classname}
"""
        if doc:
            doc_string += " *\n"
            doc_string += doc_to_doc_string(doc)
            doc_string += "\n"
        doc_string += " */"
        with open(self.current_class_target_file, "w") as f:
            _logger.info("Writing file: %s", self.current_class_target_file)
            f.write(
                """
import {{
  Dictionary,
  expandUrl,
  loadField,
  LoaderInstances,
  LoadingOptions,
  Saveable,
  ValidationException,
  prefixUrl,
  save,
  saveRelativeUri
}} from './util/Internal'
import {{ v4 as uuidv4 }} from 'uuid'
import * as Internal from './util/Internal'

{docstring}
export class {cls} extends Saveable implements Internal.{current_interface} {{
  extensionFields?: Internal.Dictionary<any>
""".format(
                    cls=cls,
                    current_interface=self.current_interface,
                    docstring=doc_string,
                )
            )
        self.current_constructor_signature.write(
            "\n" + "\n" + "  constructor ({loadingOptions, extensionFields"
        )
        self.current_constructor_body.write(
            """
    super(loadingOptions)
    this.extensionFields = extensionFields ?? {}
"""
        )
        self.current_loader.write(
            """
  /**
   * Used to construct instances of {{@link {cls} }}.
   *
   * @param __doc                           Document fragment to load this record object from.
   * @param baseuri                         Base URI to generate child document IDs against.
   * @param loadingOptions                  Context for loading URIs and populating objects.
   * @param docRoot                         ID at this position in the document (if available)
   * @returns                               An instance of {{@link {cls} }}
   * @throws {{@link ValidationException}}    If the document fragment is not a
   *                                        {{@link Dictionary}} or validation of fields fails.
   */
  static override async fromDoc (__doc: any, baseuri: string, loadingOptions: LoadingOptions,
    docRoot?: string): Promise<Saveable> {{
    const _doc = Object.assign({{}}, __doc)
    const __errors: ValidationException[] = []
            """.format(
                cls=cls
            )
        )

        self.current_serializer.write(
            """
  save (top: boolean = false, baseUrl: string = '', relativeUris: boolean = true)
  : Dictionary<any> {
    const r: Dictionary<any> = {}
    for (const ef in this.extensionFields) {
      r[prefixUrl(ef, this.loadingOptions.vocab)] = this.extensionFields.ef
    }
"""
        )

    def end_class(self, classname: str, field_names: List[str]) -> None:
        """Signal that we are done with this class."""
        with open(self.current_interface_target_file, "a") as f:
            f.write("}")
        if self.current_class_is_abstract:
            return

        self.current_constructor_signature.write(
            f"}} : {{loadingOptions?: LoadingOptions}} & Internal.{self.current_interface}) {{"
        )
        self.current_constructor_body.write("  }\n")
        self.current_loader.write(
            """
    const extensionFields: Dictionary<any> = {{}}
    for (const [key, value] of Object.entries(_doc)) {{
      if (!{classname}.attr.has(key)) {{
        if ((key as string).includes(':')) {{
          const ex = expandUrl(key, '', loadingOptions, false, false)
          extensionFields[ex] = value
        }} else {{
          __errors.push(
            new ValidationException(`invalid field ${{key as string}}, \\
            expected one of: {fields}`)
          )
          break
        }}
      }}
    }}

    if (__errors.length > 0) {{
      throw new ValidationException("Trying '{classname}'", __errors)
    }}

    const schema = new {classname}({{
      extensionFields: extensionFields,
      loadingOptions: loadingOptions,
      """.format(
                classname=self.current_class,
                fields=",".join(["\\`" + f + "\\`" for f in field_names]),
            )
        )
        self.current_loader.write(
            ",\n      ".join(
                self.safe_name(f) + ": " + self.safe_name(f) for f in field_names
            )
            + "\n    })"
        )
        self.current_loader.write(
            """
    return schema
  }
        """
        )
        self.current_serializer.write(
            """
    if (top) {
      if (this.loadingOptions.namespaces != null) {
        r.$namespaces = this.loadingOptions.namespaces
      }
      if (this.loadingOptions.schemas != null) {
        r.$schemas = this.loadingOptions.schemas
      }
    }
    return r
  }
            """
        )
        with open(
            self.current_class_target_file,
            "a",
        ) as f:
            f.write(self.current_constructor_signature.getvalue())
            f.write(self.current_constructor_body.getvalue())
            f.write(self.current_loader.getvalue())
            f.write(self.current_serializer.getvalue())
            f.write(
                "\n"
                + "  static attr: Set<string> = new Set(["
                + ",".join(["'" + shortname(f) + "'" for f in field_names])
                + "])"
            )
            f.write(
                """
}
"""
            )

    def type_loader(
        self, type_declaration: Union[List[Any], Dict[str, Any], str]
    ) -> TypeDef:
        """Parse the given type declaration and declare its components."""
        if isinstance(type_declaration, MutableSequence):
            sub_types = [self.type_loader(i) for i in type_declaration]
            sub_names: List[str] = list(dict.fromkeys([i.name for i in sub_types]))
            sub_instance_types: List[str] = list(
                dict.fromkeys(
                    [i.instance_type for i in sub_types if i.instance_type is not None]
                )
            )
            return self.declare_type(
                TypeDef(
                    "unionOf{}".format("Or".join(sub_names)),
                    "new _UnionLoader([{}])".format(", ".join(sub_names)),
                    instance_type=" | ".join(sub_instance_types),
                )
            )
        if isinstance(type_declaration, MutableMapping):
            if type_declaration["type"] in (
                "array",
                "https://w3id.org/cwl/salad#array",
            ):
                i = self.type_loader(type_declaration["items"])
                return self.declare_type(
                    TypeDef(
                        f"arrayOf{i.name}",
                        f"new _ArrayLoader([{i.name}])",
                        instance_type=f"Array<{i.instance_type}>",
                    )
                )
            if type_declaration["type"] in ("enum", "https://w3id.org/cwl/salad#enum"):
                return self.type_loader_enum(type_declaration)

            if type_declaration["type"] in (
                "record",
                "https://w3id.org/cwl/salad#record",
            ):
                return self.declare_type(
                    TypeDef(
                        self.safe_name(type_declaration["name"]) + "Loader",
                        "new _RecordLoader({}.fromDoc)".format(
                            self.safe_name(type_declaration["name"]),
                        ),
                        instance_type="Internal."
                        + self.safe_name(type_declaration["name"]),
                        abstract=type_declaration.get("abstract", False),
                    )
                )
            raise SchemaException("wft {}".format(type_declaration["type"]))

        if type_declaration in prims:
            return prims[type_declaration]

        if type_declaration in ("Expression", "https://w3id.org/cwl/cwl#Expression"):
            return self.declare_type(
                TypeDef(
                    self.safe_name(type_declaration) + "Loader",
                    "new _ExpressionLoader()",
                    instance_type="string",
                )
            )
        return self.collected_types[self.safe_name(type_declaration) + "Loader"]

    def type_loader_enum(self, type_declaration: Dict[str, Any]) -> TypeDef:
        for sym in type_declaration["symbols"]:
            self.add_vocab(shortname(sym), sym)
        enum_name = self.safe_name(type_declaration["name"])
        enum_module_name = enum_name
        enum_path = self.main_src_dir / f"{enum_module_name}.ts"
        self.modules.add(enum_module_name)
        self.record_types.add(enum_name)
        with open(enum_path, "w") as f:
            _logger.info("Writing file: %s", enum_path)
            f.write(
                """
export enum {enum_name} {{
""".format(
                    enum_name=enum_name
                )
            )
            for sym in type_declaration["symbols"]:
                val = self.safe_name(sym)
                const = self.safe_name(sym).replace("-", "_").replace(".", "_").upper()
                f.write(f"""  {const}='{val}',\n""")
            f.write(
                """}
"""
            )
        return self.declare_type(
            TypeDef(
                instance_type="Internal." + enum_name,
                name=self.safe_name(type_declaration["name"]) + "Loader",
                init=f"new _EnumLoader((Object.keys({enum_name}) as Array<keyof typeof "
                f"{enum_name}>).map(key => {enum_name}[key]))",
            )
        )

    def declare_field(
        self,
        name: str,
        fieldtype: TypeDef,
        doc: Optional[str],
        optional: bool,
    ) -> None:
        """Output the code to load the given field."""
        safename = self.safe_name(name)
        fieldname = shortname(name)
        self.current_fieldtypes[safename] = fieldtype
        if (
            fieldtype.instance_type is not None
            and "undefined" in fieldtype.instance_type
        ):
            optionalstring = "?"
        else:
            optionalstring = ""

        with open(self.current_interface_target_file, "a") as f:
            if doc:
                f.write(
                    """
  /**
{doc_str}
   */
""".format(
                        doc_str=doc_to_doc_string(doc, indent_level=1)
                    )
                )
            f.write(
                "  {safename}{optionalstring}: {type}\n".format(
                    safename=safename,
                    type=fieldtype.instance_type,
                    optionalstring=optionalstring,
                )
            )

        if self.current_class_is_abstract:
            return

        with open(self.current_class_target_file, "a") as f:
            if doc:
                f.write(
                    """
  /**
{doc_str}
   */
""".format(
                        doc_str=doc_to_doc_string(doc, indent_level=1)
                    )
                )
            f.write(
                "  {safename}{optionalstring}: {type}\n".format(
                    safename=safename,
                    type=fieldtype.instance_type,
                    optionalstring=optionalstring,
                )
            )
        if fieldname == "class":
            self.current_constructor_signature.write(
                ", {safename} = {type}.{val}".format(
                    safename=safename,
                    type=fieldtype.instance_type,
                    val=self.current_class.replace("-", "_").replace(".", "_").upper(),
                )
            )
        else:
            self.current_constructor_signature.write(
                ", {safename}".format(
                    safename=safename,
                )
            )
        self.current_constructor_body.write(
            "    this.{safeName} = {safeName}\n".format(safeName=safename)
        )

        self.current_loader.write(
            """
    let {safename}""".format(
                safename=safename
            )
        )
        if optional:
            self.current_loader.write(
                """
    if ('{fieldname}' in _doc) {{""".format(
                    fieldname=fieldname
                )
            )
            spc = "  "
        else:
            spc = ""

        self.current_loader.write(
            """
{spc}    try {{
{spc}      {safename} = await loadField(_doc.{fieldname}, LoaderInstances.{fieldtype},
{spc}        baseuri, loadingOptions)
{spc}    }} catch (e) {{
{spc}      if (e instanceof ValidationException) {{
{spc}        __errors.push(
{spc}          new ValidationException('the `{fieldname}` field is not valid because: ', [e])
{spc}        )
{spc}      }} else {{
{spc}        throw e
{spc}      }}
{spc}    }}
""".format(
                safename=safename,
                fieldname=fieldname,
                fieldtype=fieldtype.name,
                spc=spc,
            )
        )
        if optional:
            self.current_loader.write("    }\n")

        if name == self.idfield or not self.idfield:
            baseurl = "baseUrl"
        else:
            baseurl = f"this.{self.safe_name(self.idfield)}"

        if fieldtype.is_uri:
            self.current_serializer.write(
                """
    if (this.{safename} != null) {{
      const u = saveRelativeUri(this.{safename}, {base_url}, {scoped_id},
                                relativeUris, {ref_scope})
      if (u != null) {{
        r.{fieldname} = u
      }}
    }}
                """.format(
                    safename=self.safe_name(name),
                    fieldname=shortname(name).strip(),
                    base_url=baseurl,
                    scoped_id=self.to_typescript(fieldtype.scoped_id),
                    ref_scope=self.to_typescript(fieldtype.ref_scope),
                )
            )
        else:
            self.current_serializer.write(
                """
    if (this.{safename} != null) {{
      r.{fieldname} = save(this.{safename}, false, {base_url}, relativeUris)
    }}
                """.format(
                    safename=self.safe_name(name),
                    fieldname=shortname(name).strip(),
                    base_url=baseurl,
                )
            )

    def declare_id_field(
        self,
        name: str,
        fieldtype: TypeDef,
        doc: str,
        optional: bool,
        subscope: Optional[str],
    ) -> None:
        """Output the code to handle the given ID field."""
        self.declare_field(name, fieldtype, doc, True)
        if optional:
            opt = """{safename} = "_" + uuidv4()""".format(
                safename=self.safe_name(name)
            )
        else:
            opt = """throw new ValidationException("Missing {fieldname}")""".format(
                fieldname=shortname(name)
            )

        if subscope is not None:
            name = name + subscope

        self.current_loader.write(
            """
    const original{safename}IsUndefined = ({safename} === undefined)
    if (original{safename}IsUndefined ) {{
      if (docRoot != null) {{
        {safename} = docRoot
      }} else {{
        {opt}
      }}
    }} else {{
      baseuri = {safename} as string
    }}
            """.format(
                safename=self.safe_name(name), opt=opt
            )
        )

    def to_typescript(self, val: Any) -> Any:
        """Convert a Python keyword to a TypeScript keyword."""
        if val is True:
            return "true"
        elif val is None:
            return "undefined"
        elif val is False:
            return "false"
        return val

    def uri_loader(
        self,
        inner: TypeDef,
        scoped_id: bool,
        vocab_term: bool,
        ref_scope: Optional[int],
    ) -> TypeDef:
        """Construct the TypeDef for the given URI loader."""
        instance_type = inner.instance_type or "any"
        return self.declare_type(
            TypeDef(
                f"uri{inner.name}{scoped_id}{vocab_term}{ref_scope}",
                "new _URILoader({}, {}, {}, {})".format(
                    inner.name,
                    self.to_typescript(scoped_id),
                    self.to_typescript(vocab_term),
                    self.to_typescript(ref_scope),
                ),
                is_uri=True,
                scoped_id=scoped_id,
                ref_scope=ref_scope,
                instance_type=instance_type,
            )
        )

    def idmap_loader(
        self, field: str, inner: TypeDef, map_subject: str, map_predicate: Optional[str]
    ) -> TypeDef:
        """Construct the TypeDef for the given mapped ID loader."""
        instance_type = inner.instance_type or "any"
        return self.declare_type(
            TypeDef(
                f"idmap{self.safe_name(field)}{inner.name}",
                "new _IdMapLoader({}, '{}', '{}')".format(
                    inner.name, map_subject, map_predicate
                ),
                instance_type=instance_type,
            )
        )

    def typedsl_loader(self, inner: TypeDef, ref_scope: Optional[int]) -> TypeDef:
        """Construct the TypeDef for the given DSL loader."""
        instance_type = inner.instance_type or "any"
        return self.declare_type(
            TypeDef(
                f"typedsl{self.safe_name(inner.name)}{ref_scope}",
                f"new _TypeDSLLoader({self.safe_name(inner.name)}, {ref_scope})",
                instance_type=instance_type,
            )
        )

    def epilogue(self, root_loader: TypeDef) -> None:
        """Trigger to generate the epilouge code."""
        pd = "This project contains TypeScript objects and utilities "
        pd = pd + ' auto-generated by <a href=\\"https://github.com/'
        pd = pd + 'common-workflow-language/schema_salad\\">Schema Salad</a>'
        pd = pd + " for parsing documents corresponding to the "
        pd = pd + str(self.base_uri) + " schema."
        generated_class_imports = ",\n  ".join(self.record_types)
        template_vars: MutableMapping[str, str] = dict(
            project_name=self.package,
            version="0.0.1-SNAPSHOT",
            project_description=pd,
            license_name="Apache License, Version 2.0",
            generated_class_imports=generated_class_imports,
        )

        def template_from_resource(resource: str) -> string.Template:
            template_str = pkg_resources.resource_string(
                __name__, f"typescript/{resource}"
            ).decode("utf-8")
            template = string.Template(template_str)
            return template

        def expand_resource_template_to(resource: str, path: Path) -> None:
            template = template_from_resource(resource)
            src = template.safe_substitute(template_vars)
            _ensure_directory_and_write(path, src)

        expand_resource_template_to("package.json", self.target_dir / "package.json")
        expand_resource_template_to(".gitignore", self.target_dir / ".gitignore")
        expand_resource_template_to("LICENSE", self.target_dir / "LICENSE")
        expand_resource_template_to("tsconfig.json", self.target_dir / "tsconfig.json")
        expand_resource_template_to("index.ts", self.main_src_dir / "index.ts")

        vocab = ",\n  ".join(
            f"""'{k}': '{self.vocab[k]}'""" for k in sorted(self.vocab.keys())
        )
        rvocab = ",\n  ".join(
            f"""'{self.vocab[k]}': '{k}'""" for k in sorted(self.vocab.keys())
        )

        loader_instances = ""
        for _, collected_type in self.collected_types.items():
            if not collected_type.abstract:
                loader_instances += "export const {} = {};\n".format(
                    collected_type.name, collected_type.init
                )

        internal_module_exports = "\n".join(
            f"export * from '../{f}'" for f in self.modules
        )

        example_tests = ""
        if self.examples:
            _safe_makedirs(self.test_resources_dir)
            utils_resources = self.test_resources_dir / "examples"
            if os.path.exists(utils_resources):
                shutil.rmtree(utils_resources)
            shutil.copytree(self.examples, utils_resources)
            for example_name in os.listdir(self.examples):
                if example_name.startswith("valid"):
                    basename = os.path.basename(example_name).rsplit(".", 1)[0]
                    example_tests += """
    it('{basename}', async () => {{
        await loadDocument(__dirname + '/data/examples/{example_name}')
    }})
    it('{basename} by string', async () => {{
        let doc = fs.readFileSync(__dirname + '/data/examples/{example_name}').toString()
        await loadDocumentByString(doc, url.pathToFileURL(__dirname +
            '/data/examples/').toString())
    }})""".format(
                        basename=basename.replace("-", "_").replace(".", "_"),
                        example_name=example_name,
                    )

        template_args: MutableMapping[str, str] = dict(
            internal_module_exports=internal_module_exports,
            loader_instances=loader_instances,
            generated_class_imports=generated_class_imports,
            vocab=vocab,
            rvocab=rvocab,
            root_loader=root_loader.name,
            root_loader_type=root_loader.instance_type or "any",
            tests=example_tests,
        )

        util_src_dirs = {
            "util": self.main_src_dir / "util",
            "test": self.main_src_dir / "test",
        }

        def copy_utils_recursive(util_src: str, util_target: Path) -> None:
            for util in pkg_resources.resource_listdir(
                __name__, f"typescript/{util_src}"
            ):
                template_path = os.path.join(util_src, util)
                if pkg_resources.resource_isdir(
                    __name__, f"typescript/{template_path}"
                ):
                    copy_utils_recursive(
                        os.path.join(util_src, util), util_target / util
                    )
                    continue
                src_path = util_target / util
                src_template = template_from_resource(template_path)
                src = src_template.safe_substitute(template_args)
                _ensure_directory_and_write(src_path, src)

        for (util_src, util_target) in util_src_dirs.items():
            copy_utils_recursive(util_src, util_target)

    def secondaryfilesdsl_loader(self, inner: TypeDef) -> TypeDef:
        """Construct the TypeDef for secondary files."""
        instance_type = inner.instance_type or "any"
        return self.declare_type(
            TypeDef(
                f"secondaryfilesdsl{inner.name}",
                f"new _SecondaryDSLLoader({inner.name})",
                instance_type=instance_type,
            )
        )



import json
import os
import sys
from io import BufferedWriter
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Mapping,
    MutableSequence,
    Tuple,
    TypeVar,
    Union,
)

import requests
from rdflib.graph import Graph
from ruamel.yaml.comments import CommentedMap, CommentedSeq
from ruamel.yaml.constructor import RoundTripConstructor
from ruamel.yaml.main import YAML

if TYPE_CHECKING:
    from .fetcher import Fetcher

ContextType = Dict[str, Union[Dict[str, Any], str, Iterable[str]]]
DocumentType = TypeVar("DocumentType", CommentedSeq, CommentedMap)
DocumentOrStrType = TypeVar("DocumentOrStrType", CommentedSeq, CommentedMap, str)
FieldType = TypeVar("FieldType", str, CommentedSeq, CommentedMap)
ResolveType = Union[int, float, str, CommentedMap, CommentedSeq, None]
ResolvedRefType = Tuple[ResolveType, CommentedMap]
IdxResultType = Union[CommentedMap, CommentedSeq, str, None]
IdxType = Dict[str, IdxResultType]
CacheType = Dict[str, Union[str, Graph, bool]]
FetcherCallableType = Callable[[CacheType, requests.sessions.Session], "Fetcher"]
AttachmentsType = Callable[[Union[CommentedMap, CommentedSeq]], bool]


def add_dictlist(di, key, val):  # type: (Dict[Any, Any], Any, Any) -> None
    if key not in di:
        di[key] = []
    di[key].append(val)


def aslist(thing):  # type: (Any) -> MutableSequence[Any]
    """
    Convenience function to wrap single items and lists.

    Return lists unchanged.
    """

    if isinstance(thing, MutableSequence):
        return thing
    else:
        return [thing]


# http://rightfootin.blogspot.com/2006/09/more-on-python-flatten.html


def flatten(thing, ltypes=(list, tuple)):
    # type: (Any, Any) -> Any
    if thing is None:
        return []
    if not isinstance(thing, ltypes):
        return [thing]

    ltype = type(thing)
    lst = list(thing)
    i = 0
    while i < len(lst):
        while isinstance(lst[i], ltypes):
            if not lst[i]:
                lst.pop(i)
                i -= 1
                break
            else:
                lst[i : i + 1] = lst[i]
        i += 1
    return ltype(lst)


# Check if we are on windows OS
def onWindows():
    # type: () -> (bool)
    return os.name == "nt"


def convert_to_dict(j4):  # type: (Any) -> Any
    if isinstance(j4, Mapping):
        return {k: convert_to_dict(v) for k, v in j4.items()}
    elif isinstance(j4, MutableSequence):
        return [convert_to_dict(v) for v in j4]
    else:
        return j4


def json_dump(
    obj,  # type: Any
    fp,  # type: IO[str]
    **kwargs  # type: Any
):  # type: (...) -> None
    """Force use of unicode."""
    json.dump(convert_to_dict(obj), fp, **kwargs)


def json_dumps(
    obj,  # type: Any
    **kwargs  # type: Any
):  # type: (...) -> str
    """Force use of unicode."""
    return json.dumps(convert_to_dict(obj), **kwargs)


def stdout() -> BufferedWriter:
    """Build a replacement for sys.stdout that allow for writing binary data."""
    return os.fdopen(sys.stdout.fileno(), "wb", closefd=False)


class _RoundTripNoTimeStampConstructor(RoundTripConstructor):
    def construct_yaml_timestamp(self: Any, node: Any, values: Any = None) -> Any:
        return node.value


_RoundTripNoTimeStampConstructor.add_constructor(
    "tag:yaml.org,2002:timestamp",
    _RoundTripNoTimeStampConstructor.construct_yaml_timestamp,
)


def yaml_no_ts() -> YAML:
    """
    Get a YAML loader that won't parse timestamps into datetime objects.

    Such datetime objects can't be easily dumped into JSON.
    """
    yaml = YAML(typ="rt")
    yaml.preserve_quotes = True  # type: ignore
    yaml.Constructor = _RoundTripNoTimeStampConstructor
    return yaml



class Event:
    def __init__(self, **kw) -> None: ...

class Dispatcher:
    def set_map(self, amap): ...
    def get_map(self): ...
    def subscribe(self, event_type, handler): ...
    def dispatch(self, event) -> None: ...



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class DCAT(DefinedNamespace):
    accessURL: URIRef
    bbox: URIRef
    byteSize: URIRef
    centroid: URIRef
    compressFormat: URIRef
    contactPoint: URIRef
    dataset: URIRef
    distribution: URIRef
    downloadURL: URIRef
    endDate: URIRef
    keyword: URIRef
    landingPage: URIRef
    mediaType: URIRef
    packageFormat: URIRef
    record: URIRef
    startDate: URIRef
    theme: URIRef
    themeTaxonomy: URIRef
    Catalog: URIRef
    CatalogRecord: URIRef
    Dataset: URIRef
    Distribution: URIRef
    DataService: URIRef
    Relationship: URIRef
    Resource: URIRef
    Role: URIRef
    spatialResolutionInMeters: URIRef
    temporalResolution: URIRef
    accessService: URIRef
    catalog: URIRef
    endpointDescription: URIRef
    endpointURL: URIRef
    hadRole: URIRef
    qualifiedRelation: URIRef
    servesDataset: URIRef
    service: URIRef



from io import BufferedIOBase
from typing import Any, Optional, Union

class Processor:
    def __init__(self, graph) -> None: ...
    def query(
        self, strOrQuery, initBindings=..., initNs=..., DEBUG: bool = ...
    ) -> None: ...

class UpdateProcessor:
    def __init__(self, graph) -> None: ...
    def update(self, strOrQuery, initBindings=..., initNs=...) -> None: ...

class ResultException(Exception): ...

class EncodeOnlyUnicode:
    def __init__(self, stream) -> None: ...
    def write(self, arg) -> None: ...
    def __getattr__(self, name): ...

class ResultRow(tuple):
    def __new__(cls, values, labels): ...
    def __getattr__(self, name): ...
    def __getitem__(self, name): ...
    def get(self, name, default: Any | None = ...): ...
    def asdict(self): ...

class Result:
    type: Any
    vars: Any
    askAnswer: Any
    graph: Any
    def __init__(self, type_) -> None: ...
    bindings: Any
    @staticmethod
    def parse(
        source: Any | None = ...,
        format: Any | None = ...,
        content_type: Any | None = ...,
        **kwargs
    ): ...
    def serialize(
        self,
        destination: Optional[Union[str, BufferedIOBase]] = ...,
        encoding: str = ...,
        format: str = ...,
        **args
    ) -> Optional[bytes]: ...
    def __len__(self): ...
    def __bool__(self): ...
    def __iter__(self): ...
    def __getattr__(self, name): ...
    def __eq__(self, other): ...

class ResultParser:
    def __init__(self) -> None: ...
    def parse(self, source, **kwargs) -> None: ...

class ResultSerializer:
    result: Any
    def __init__(self, result) -> None: ...
    def serialize(self, stream, encoding: str = ..., **kwargs) -> None: ...



# Stubs for cachecontrol.cache (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

class BaseCache:
    def get(self, key): ...
    def set(self, key, value): ...
    def delete(self, key): ...
    def close(self): ...

class DictCache(BaseCache):
    lock = ...  # type: Any
    data = ...  # type: Any
    def __init__(self, init_dict=None): ...
    def get(self, key): ...
    def set(self, key, value): ...
    def delete(self, key): ...



# Stubs for cachecontrol.wrapper (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

from .adapter import CacheControlAdapter
from .cache import DictCache

def CacheControl(
    sess: Any,
    cache: Optional[Any] = ...,
    cache_etags: bool = ...,
    serializer: Optional[Any] = ...,
    heuristic: Optional[Any] = ...,
): ...



from typing import Any, Type, TypeVar

from rdflib.exceptions import Error

class PluginException(Error): ...

class Plugin:
    name: Any
    kind: Any
    module_path: Any
    class_name: Any
    def __init__(self, name, kind, module_path, class_name) -> None: ...
    def getClass(self): ...

class PKGPlugin(Plugin):
    name: Any
    kind: Any
    ep: Any
    def __init__(self, name, kind, ep) -> None: ...
    def getClass(self): ...

def register(name: str, kind, module_path, class_name): ...

PluginT = TypeVar("PluginT")

def get(name: str, kind: Type[PluginT]) -> Type[PluginT]: ...
def plugins(name: Any | None = ..., kind: Any | None = ...) -> None: ...



import pickle
from typing import Any

from rdflib.events import Event

Pickler = pickle.Pickler
Unpickler = pickle.Unpickler
UnpicklingError = pickle.UnpicklingError

class StoreCreatedEvent(Event): ...
class TripleAddedEvent(Event): ...
class TripleRemovedEvent(Event): ...

class NodePickler:
    def __init__(self) -> None: ...
    def register(self, object, id) -> None: ...
    def loads(self, s): ...
    def dumps(self, obj, protocol: Any | None = ..., bin: Any | None = ...): ...

class Store:
    context_aware: bool
    formula_aware: bool
    transaction_aware: bool
    graph_aware: bool
    dispatcher: Any
    def __init__(
        self, configuration: Any | None = ..., identifier: Any | None = ...
    ) -> None: ...
    node_pickler: Any
    def create(self, configuration) -> None: ...
    def open(self, configuration, create: bool = ...): ...
    def close(self, commit_pending_transaction: bool = ...) -> None: ...
    def destroy(self, configuration) -> None: ...
    def gc(self) -> None: ...
    def add(self, triple, context, quoted: bool = ...) -> None: ...
    def addN(self, quads) -> None: ...
    def remove(self, triple, context: Any | None = ...) -> None: ...
    def triples_choices(self, triple, context: Any | None = ...) -> None: ...
    def triples(self, triple_pattern, context: Any | None = ...) -> None: ...
    def __len__(self, context: Any | None = ...) -> None: ...
    def contexts(self, triple: Any | None = ...) -> None: ...
    def query(self, query, initNs, initBindings, queryGraph, **kwargs) -> None: ...
    def update(self, update, initNs, initBindings, queryGraph, **kwargs) -> None: ...
    def bind(self, prefix, namespace) -> None: ...
    def prefix(self, namespace) -> None: ...
    def namespace(self, prefix) -> None: ...
    def namespaces(self) -> None: ...
    def commit(self) -> None: ...
    def rollback(self) -> None: ...
    def add_graph(self, graph) -> None: ...
    def remove_graph(self, graph) -> None: ...



# Stubs for cachecontrol.serialize (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from .compat import HTTPResponse as HTTPResponse
from .compat import pickle as pickle
from .compat import str as str

class Serializer:
    def dumps(self, request, response, body=None): ...
    def loads(self, request, data): ...
    def prepare_response(self, request, cached): ...



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class CSVW(DefinedNamespace):
    aboutUrl: URIRef
    base: URIRef
    column: URIRef
    columnReference: URIRef
    commentPrefix: URIRef
    datatype: URIRef
    decimalChar: URIRef
    default: URIRef
    delimiter: URIRef
    describes: URIRef
    dialect: URIRef
    doubleQuote: URIRef
    encoding: URIRef
    foreignKey: URIRef
    format: URIRef
    groupChar: URIRef
    header: URIRef
    headerRowCount: URIRef
    lang: URIRef
    length: URIRef
    lineTerminators: URIRef
    maxExclusive: URIRef
    maxInclusive: URIRef
    maxLength: URIRef
    minExclusive: URIRef
    minInclusive: URIRef
    minLength: URIRef
    name: URIRef
    note: URIRef
    null: URIRef
    ordered: URIRef
    pattern: URIRef
    primaryKey: URIRef
    propertyUrl: URIRef
    quoteChar: URIRef
    reference: URIRef
    referencedRow: URIRef
    required: URIRef
    resource: URIRef
    row: URIRef
    rowTitle: URIRef
    rownum: URIRef
    schemaReference: URIRef
    scriptFormat: URIRef
    separator: URIRef
    skipBlankRows: URIRef
    skipColumns: URIRef
    skipInitialSpace: URIRef
    skipRows: URIRef
    source: URIRef
    suppressOutput: URIRef
    table: URIRef
    tableDirection: URIRef
    tableSchema: URIRef
    targetFormat: URIRef
    textDirection: URIRef
    title: URIRef
    transformations: URIRef
    trim: URIRef
    url: URIRef
    valueUrl: URIRef
    virtual: URIRef
    Cell: URIRef
    Column: URIRef
    Datatype: URIRef
    Dialect: URIRef
    Direction: URIRef
    ForeignKey: URIRef
    NumericFormat: URIRef
    Row: URIRef
    Schema: URIRef
    Table: URIRef
    TableGroup: URIRef
    TableReference: URIRef
    Transformation: URIRef
    JSON: URIRef
    uriTemplate: URIRef
    auto: URIRef
    inherit: URIRef
    ltr: URIRef
    rtl: URIRef
    csvEncodedTabularData: URIRef
    tabularMetadata: URIRef



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class VOID(DefinedNamespace):
    classPartition: URIRef
    classes: URIRef
    dataDump: URIRef
    distinctObjects: URIRef
    distinctSubjects: URIRef
    documents: URIRef
    entities: URIRef
    exampleResource: URIRef
    feature: URIRef
    inDataset: URIRef
    linkPredicate: URIRef
    objectsTarget: URIRef
    openSearchDescription: URIRef
    properties: URIRef
    property: URIRef
    propertyPartition: URIRef
    rootResource: URIRef
    sparqlEndpoint: URIRef
    subjectsTarget: URIRef
    subset: URIRef
    target: URIRef
    triples: URIRef
    uriLookupEndpoint: URIRef
    uriRegexPattern: URIRef
    uriSpace: URIRef
    vocabulary: URIRef
    Dataset: URIRef
    DatasetDescription: URIRef
    Linkset: URIRef
    TechnicalFeature: URIRef
